#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import print_function
from bitdust_forks.CodernityDB.tree_index import TreeBasedIndex
import struct
import os

import inspect
from functools import wraps
import json
from six.moves import range
from io import open


class DebugTreeBasedIndex(TreeBasedIndex):

    def __init__(self, *args, **kwargs):
        super(DebugTreeBasedIndex, self).__init__(*args, **kwargs)

    def print_tree(self):
        print('-----CURRENT TREE-----')
        print(self.root_flag)

        if self.root_flag == 'l':
            print('---ROOT---')
            self._print_leaf_data(self.data_start)
            return
        else:
            print('---ROOT---')
            self._print_node_data(self.data_start)
            nr_of_el, children_flag = self._read_node_nr_of_elements_and_children_flag(
                self.data_start)
            nodes = []
            for index in range(nr_of_el):
                l_pointer, key, r_pointer = self._read_single_node_key(
                    self.data_start, index)
                nodes.append(l_pointer)
            nodes.append(r_pointer)
            print('ROOT NODES', nodes)
            while children_flag == 'n':
                self._print_level(nodes, 'n')
                new_nodes = []
                for node in nodes:
                    nr_of_el, children_flag = \
                        self._read_node_nr_of_elements_and_children_flag(node)
                    for index in range(nr_of_el):
                        l_pointer, key, r_pointer = self._read_single_node_key(
                            node, index)
                        new_nodes.append(l_pointer)
                    new_nodes.append(r_pointer)
                nodes = new_nodes
            self._print_level(nodes, 'l')

    def _print_level(self, nodes, flag):
        print('---NEXT LVL---')
        if flag == 'n':
            for node in nodes:
                self._print_node_data(node)
        elif flag == 'l':
            for node in nodes:
                self._print_leaf_data(node)

    def _print_leaf_data(self, leaf_start_position):
        print('printing data of leaf at', leaf_start_position)
        nr_of_elements = self._read_leaf_nr_of_elements(leaf_start_position)
        self.buckets.seek(leaf_start_position)
        data = self.buckets.read(self.leaf_heading_size +
                                 nr_of_elements * self.single_leaf_record_size)
        leaf = struct.unpack('<' + self.leaf_heading_format +
                             nr_of_elements * self.single_leaf_record_format, data)
        print(leaf)
        print()

    def _print_node_data(self, node_start_position):
        print('printing data of node at', node_start_position)
        nr_of_elements = self._read_node_nr_of_elements_and_children_flag(
            node_start_position)[0]
        self.buckets.seek(node_start_position)
        data = self.buckets.read(self.node_heading_size + self.pointer_size
                                 + nr_of_elements * (self.key_size + self.pointer_size))
        node = struct.unpack('<' + self.node_heading_format + self.pointer_format
                             + nr_of_elements * (
                                 self.key_format + self.pointer_format),
                             data)
        print(node)
        print()
# ------------------>


def database_step_by_step(db_obj, path=None):

    if not path:
        # ugly for multiplatform support....
        p = db_obj.path
        p1 = os.path.split(p)
        p2 = os.path.split(p1[0])
        p3 = '_'.join([p2[1], 'operation_logger.log'])
        path = os.path.join(os.path.split(p2[0])[0], p3)
    f_obj = open(path, 'wb')

    __stack = []  # inspect.stack() is not working on pytest etc

    def remove_from_stack(name):
        for i in range(len(__stack)):
            if __stack[-i] == name:
                __stack.pop(-i)

    def __dumper(f):
        @wraps(f)
        def __inner(*args, **kwargs):
            funct_name = f.__name__
            if funct_name == 'count':
                name = args[0].__name__
                meth_args = (name,) + args[1:]
            elif funct_name in ('reindex_index', 'compact_index'):
                name = args[0].name
                meth_args = (name,) + args[1:]
            else:
                meth_args = args
            kwargs_copy = kwargs.copy()
            res = None
            __stack.append(funct_name)
            if funct_name == 'insert':
                try:
                    res = f(*args, **kwargs)
                except:
                    packed = json.dumps((funct_name,
                                         meth_args, kwargs_copy, None))
                    f_obj.write('%s\n' % packed)
                    f_obj.flush()
                    raise
                else:
                    packed = json.dumps((funct_name,
                                         meth_args, kwargs_copy, res))
                f_obj.write('%s\n' % packed)
                f_obj.flush()
            else:
                if funct_name == 'get':
                    for curr in __stack:
                        if ('delete' in curr or 'update' in curr) and not curr.startswith('test'):
                            remove_from_stack(funct_name)
                            return f(*args, **kwargs)
                packed = json.dumps((funct_name, meth_args, kwargs_copy))
                f_obj.write('%s\n' % packed)
                f_obj.flush()
                res = f(*args, **kwargs)
            remove_from_stack(funct_name)
            return res
        return __inner

    for meth_name, meth_f in inspect.getmembers(db_obj, predicate=inspect.ismethod):
        if not meth_name.startswith('_'):
            setattr(db_obj, meth_name, __dumper(meth_f))

    setattr(db_obj, 'operation_logger', f_obj)


def database_from_steps(db_obj, path):
    # db_obj.insert=lambda data : insert_for_debug(db_obj, data)
    with open(path, 'rb') as f_obj:
        for current in f_obj:
            line = json.loads(current[:-1])
            if line[0] == 'count':
                obj = getattr(db_obj, line[1][0])
                line[1] = [obj] + line[1][1:]
            name = line[0]
            if name == 'insert':
                try:
                    line[1][0].pop('_rev')
                except:
                    pass
            elif name in ('delete', 'update'):
                el = db_obj.get('id', line[1][0]['_id'])
                line[1][0]['_rev'] = el['_rev']
#                print 'FROM STEPS doing', line
            meth = getattr(db_obj, line[0], None)
            if not meth:
                raise Exception("Method = `%s` not found" % line[0])

            meth(*line[1], **line[2])


# def insert_for_debug(self, data):
#
#    _rev = data['_rev']
#
#    if not '_id' in data:
#        _id = uuid4().hex
#    else:
#        _id = data['_id']
#    data['_id'] = _id
#    try:
#        _id = bytes(_id)
#    except:
#        raise DatabaseException("`_id` must be valid bytes object")
#    self._insert_indexes(_id, _rev, data)
#    ret = {'_id': _id, '_rev': _rev}
#    data.update(ret)
#    return ret



#!/usr/bin/env python

from __future__ import absolute_import

USERS = {'admin': 'admin', 'user': 'user', 'test': 'eW91IGFyZSBjcmF6eQo='}
"""
Twisted SSL webserver with basic authentication using plain in-memory passwords.
The first argument is the path of the directory to serve; if not provided then the current folder is used (".").

INSTALL DEPENDENCIES:
    pip install twisted
    pip install pyOpenSSL
    pip install service_identity

GENERATE SSL CERTIFICATES:
    mkdir ~/.ssl && cd ~/.ssl
    openssl genrsa > privkey.pem
    openssl req -new -x509 -key privkey.pem -out cacert.pem -days 9999

USAGE:
    Requires running as root (normal users cannot bind to ports below 1024);
    login with test_user/test_password

    sudo python twisted-web-ssl.py     # serve the current folder
    sudo python twisted-web-ssl.py /home
"""
import os
import sys

from twisted.web.static import File
from zope.interface import implementer
from twisted.python import log
from twisted.internet import reactor  # @UnresolvedImport
from twisted.web import server, resource, guard
from twisted.cred.portal import IRealm, Portal
from twisted.cred.checkers import InMemoryUsernamePasswordDatabaseDontUse
from twisted.python.log import startLogging

startLogging(sys.stdout)
home_dir = os.path.expanduser('~')

sslContext = ssl.DefaultOpenSSLContextFactory(
    os.path.join(home_dir, '.ssl/privkey.pem'),
    os.path.join(home_dir, '.ssl/cacert.pem'),
)


@implementer(IRealm)
class SimpleRealm(object):
    def __init__(self, path):
        self.path = path

    def requestAvatar(self, avatarId, mind, *interfaces):

        if resource.IResource in interfaces:
            return resource.IResource, File(self.path), lambda: None

        raise NotImplementedError()


def main(root):
    log.startLogging(sys.stdout)
    checkers = [InMemoryUsernamePasswordDatabaseDontUse(**USERS)]

    wrapper = guard.HTTPAuthSessionWrapper(Portal(SimpleRealm(root), checkers), [guard.DigestCredentialFactory('md5', 'whatever.com')])

    reactor.listenSSL(443, server.Site(resource=wrapper), contextFactory=sslContext)
    reactor.run()


if __name__ == '__main__':
    root = sys.argv[1] if len(sys.argv) > 1 else '.'
    main(root)



#!/usr/bin/python
# key_ring.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (key_ring.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: key_ring

"""

#------------------------------------------------------------------------------

from __future__ import absolute_import

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 8

#------------------------------------------------------------------------------

import os
import sys
import base64

from twisted.internet.defer import Deferred, fail

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.lib import strng
from bitdust.lib import serialization

from bitdust.main import settings

from bitdust.contacts import identitycache
from bitdust.contacts import contactsdb

from bitdust.p2p import online_status
from bitdust.p2p import p2p_service
from bitdust.p2p import commands

from bitdust.crypt import key
from bitdust.crypt import my_keys
from bitdust.crypt import encrypted

from bitdust.storage import backup_matrix

from bitdust.interface import api

from bitdust.userid import global_id
from bitdust.userid import my_id

#------------------------------------------------------------------------------

_MyKeysInSync = False

#------------------------------------------------------------------------------


def init():
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.init')
    my_keys.check_rename_my_keys()


def shutdown():
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.shutdown')


#------------------------------------------------------------------------------


def _do_request_service_keys_registry(key_id, idurl, include_private, include_signature, timeout, result):
    p2p_service.SendRequestService(
        idurl,
        'service_keys_registry',
        timeout=timeout,
        callbacks={
            commands.Ack(): lambda response, info: _on_service_keys_registry_response(response, info, key_id, idurl, include_private, include_signature, result, timeout),
            commands.Fail(): lambda response, info: result.errback(Exception('"service_keys_registry" not started on remote node')),
            None: lambda pkt_out: result.errback(Exception('timeout')),
        },
    )
    return result


def _on_service_keys_registry_response(response, info, key_id, idurl, include_private, include_signature, result, timeout):
    if not strng.to_text(response.Payload).startswith('accepted'):
        if not result.called:
            result.errback(Exception('request for "service_keys_registry" refused by remote node'))
        return None
    d = transfer_key(
        key_id,
        trusted_idurl=idurl,
        include_private=include_private,
        include_signature=include_signature,
        timeout=timeout,
        result=result,
    )
    d.addErrback(lambda *a: lg.err('transfer key failed: %s' % str(*a)))
    return None


def _on_transfer_key_response(response, info, key_id, result):
    if not response or not info:
        if not result.called:
            result.errback(Exception('timeout'))
        if _Debug:
            lg.warn('transfer failed, response timeout')
        return None
    if response.Command == commands.Ack():
        if not result.called:
            result.callback(response)
        if _Debug:
            lg.info('key %s transfer success to %s' % (key_id, response.OwnerID))
        return None
    if response.Command == commands.Fail():
        err_msg = strng.to_text(response.Payload, errors='ignore')
        if err_msg.count('key already registered'):
            # it is okay to have "Fail()" response in that case
            if not result.called:
                result.callback(response)
            if _Debug:
                lg.warn('key %s already registered on %s' % (key_id, response.OwnerID))
            return None
    if not result.called:
        result.errback(Exception(response.Payload))
    if _Debug:
        lg.warn('key transfer failed: %s' % response.Payload)
    return None


def transfer_key(key_id, trusted_idurl, include_private=False, include_signature=False, timeout=None, result=None):
    """
    Actually sending given key to remote user.
    """
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.transfer_key  %s -> %s' % (key_id, trusted_idurl))
    if not timeout:
        timeout = settings.P2PTimeOut()
    key_id = my_keys.latest_key_id(key_id)
    if not result:
        result = Deferred()
    recipient_id_obj = identitycache.FromCache(trusted_idurl)
    if not recipient_id_obj:
        lg.warn('not found "%s" in identity cache' % trusted_idurl)
        result.errback(Exception('not found "%s" in identity cache' % trusted_idurl))
        return result
    key_alias, creator_idurl = my_keys.split_key_id(key_id)
    if not key_alias or not creator_idurl:
        lg.warn('wrong key_id')
        result.errback(Exception('wrong key_id'))
        return result
    if not my_keys.is_key_registered(key_id):
        lg.warn('unknown key: "%s"' % key_id)
        result.errback(Exception('unknown key: "%s"' % key_id))
        return result
    key_object = my_keys.key_obj(key_id)
    try:
        key_json = my_keys.make_key_info(
            key_object,
            key_id=key_id,
            include_private=include_private,
            generate_signature=include_signature,
        )
    except Exception as exc:
        lg.exc()
        result.errback(exc)
        return result
    if include_signature and not my_keys.verify_key_info_signature(key_json):
        lg.err('signature verification failed after making key info: %r' % key_json)
        result.errback(Exception('signature verification failed after making key info: "%s"' % key_id))
        return result
    if _Debug:
        lg.args(_DebugLevel, key_json=key_json)
    key_data = serialization.DictToBytes(key_json, values_to_text=True)
    block = encrypted.Block(
        BackupID=key_id,
        Data=key_data,
        SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
        SessionKeyType=key.SessionKeyType(),
        # encrypt data using public key of recipient
        EncryptKey=lambda inp: recipient_id_obj.encrypt(inp),
    )
    encrypted_key_data = block.Serialize()
    p2p_service.SendKey(
        remote_idurl=recipient_id_obj.getIDURL(),
        encrypted_key_data=encrypted_key_data,
        packet_id=key_id,
        callbacks={
            commands.Ack(): lambda response, info: _on_transfer_key_response(response, info, key_id, result),
            commands.Fail(): lambda response, info: _on_transfer_key_response(response, info, key_id, result),
            None: lambda pkt_out: _on_transfer_key_response(None, None, key_id, result),
        },
        timeout=timeout,
    )
    return result


def share_key(key_id, trusted_idurl, include_private=False, include_signature=False, timeout=None):
    """
    Method to be used to send given key to one trusted user.
    Make sure remote user is identified and connected.
    Returns deferred, callback will be fired with response Ack() packet argument.
    """
    if _Debug:
        lg.args(_DebugLevel, key_id=key_id, trusted_idurl=trusted_idurl)
    if timeout is None:
        timeout = settings.P2PTimeOut()
    key_id = my_keys.latest_key_id(key_id)
    result = Deferred()
    d = online_status.ping(
        idurl=trusted_idurl,
        ack_timeout=timeout,
        channel='share_key',
        keep_alive=False,
    )
    d.addCallback(lambda ok: _do_request_service_keys_registry(
        key_id,
        trusted_idurl,
        include_private,
        include_signature,
        timeout,
        result,
    ))
    d.addErrback(result.errback)
    return result


#------------------------------------------------------------------------------


def _on_audit_public_key_response(response, info, key_id, untrusted_idurl, test_sample, result):
    try:
        response_sample = base64.b64decode(response.Payload)
    except:
        lg.exc()
        result.callback(False)
        return False
    key_alias, creator_idurl = my_keys.split_key_id(key_id)
    if creator_idurl == untrusted_idurl and key_alias == 'master':
        recipient_id_obj = identitycache.FromCache(creator_idurl)
        if not recipient_id_obj:
            lg.warn('not found "%s" in identity cache' % creator_idurl)
            result.errback(Exception('not found "%s" in identity cache' % creator_idurl))
            return result
        orig_sample = recipient_id_obj.encrypt(test_sample)
    else:
        orig_sample = my_keys.encrypt(key_id, test_sample)
    if response_sample == orig_sample:
        if _Debug:
            lg.out(_DebugLevel, 'key_ring._on_audit_public_key_response : %s on %s' % (key_id, untrusted_idurl))
        result.callback(True)
        return True
    lg.warn('key %s on %s is not OK' % (key_id, untrusted_idurl))
    result.callback(False)
    return False


def audit_public_key(key_id, untrusted_idurl, timeout=None):
    """
    Be sure remote user stores given public key.
    I also need to stores that public key in order to do such audit.
    I will send him a random string, he needs to encrypt it and send me back.
    I can compare his encrypted output with mine.
    Returns Deferred object.
    """
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.audit_public_key   testing %s from %s' % (key_id, untrusted_idurl))
    if not timeout:
        timeout = settings.P2PTimeOut()
    key_id = my_keys.latest_key_id(key_id)
    result = Deferred()
    recipient_id_obj = identitycache.FromCache(untrusted_idurl)
    if not recipient_id_obj:
        lg.warn('not found "%s" in identity cache' % untrusted_idurl)
        result.errback(Exception('not found "%s" in identity cache' % untrusted_idurl))
        return result
    key_alias, creator_idurl = my_keys.split_key_id(key_id)
    if not key_alias or not creator_idurl:
        lg.warn('wrong key_id')
        result.errback(Exception('wrong key_id'))
        return result
    if untrusted_idurl == creator_idurl and key_alias == 'master':
        lg.info('doing audit of master key (public part) of remote user')
    else:
        if not my_keys.is_key_registered(key_id):
            lg.warn('unknown key: "%s"' % key_id)
            result.errback(Exception('unknown key: "%s"' % key_id))
            return result
    public_test_sample = key.NewSessionKey(session_key_type=key.SessionKeyType())
    json_payload = {
        'key_id': key_id,
        'audit': {
            'public_sample': base64.b64encode(public_test_sample),
            'private_sample': '',
        },
    }
    raw_payload = serialization.DictToBytes(json_payload, values_to_text=True)
    block = encrypted.Block(
        BackupID=key_id,
        Data=raw_payload,
        SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
        SessionKeyType=key.SessionKeyType(),
        # encrypt data using public key of recipient
        EncryptKey=lambda inp: recipient_id_obj.encrypt(inp),
    )
    encrypted_payload = block.Serialize()
    p2p_service.SendAuditKey(
        remote_idurl=recipient_id_obj.getIDURL(),
        encrypted_payload=encrypted_payload,
        packet_id=key_id,
        timeout=timeout,
        callbacks={
            commands.Ack(): lambda response, info: _on_audit_public_key_response(response, info, key_id, untrusted_idurl, public_test_sample, result),
            commands.Fail(): lambda response, info: result.errback(Exception(response)),
            None: lambda pkt_out: result.errback(Exception('timeout')),  # timeout
        },
    )
    return result


#------------------------------------------------------------------------------


def _on_audit_private_key_response(response, info, key_id, untrusted_idurl, test_sample, result):
    try:
        response_sample = base64.b64decode(response.Payload)
    except:
        lg.exc()
        result.callback(False)
        return False
    if response_sample == test_sample:
        if _Debug:
            lg.out(_DebugLevel, 'key_ring._on_audit_private_key_response : %s on %s' % (key_id, untrusted_idurl))
        result.callback(True)
        return True
    lg.warn('key %s on %s is not OK' % (key_id, untrusted_idurl))
    result.callback(False)
    return False


def audit_private_key(key_id, untrusted_idurl, timeout=None):
    """
    Be sure remote user possess given private key.
    I need to possess the public key to be able to audit.
    I will generate a random string, encrypt it with given key public key and send encrypted string to him.
    He will decrypt and send me back original string.
    Returns Deferred object.
    """
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.audit_private_key   testing %s from %s' % (key_id, untrusted_idurl))
    if not timeout:
        timeout = settings.P2PTimeOut()
    key_id = my_keys.latest_key_id(key_id)
    result = Deferred()
    recipient_id_obj = identitycache.FromCache(untrusted_idurl)
    if not recipient_id_obj:
        lg.warn('not found "%s" in identity cache' % untrusted_idurl)
        result.errback(Exception('not found "%s" in identity cache' % untrusted_idurl))
        return result
    key_alias, creator_idurl = my_keys.split_key_id(key_id)
    if not key_alias or not creator_idurl:
        lg.warn('wrong key_id')
        result.errback(Exception('wrong key_id'))
        return result
    remote_idurl = recipient_id_obj.getIDURL()
    private_test_sample = key.NewSessionKey(session_key_type=key.SessionKeyType())
    if untrusted_idurl == creator_idurl and key_alias == 'master':
        lg.info('doing audit of master key %r for remote user %r' % (key_id, remote_idurl))
        private_test_encrypted_sample = recipient_id_obj.encrypt(private_test_sample)
    else:
        if not my_keys.is_key_registered(key_id):
            lg.warn('unknown key: "%s"' % key_id)
            result.errback(Exception('unknown key: "%s"' % key_id))
            return result
        lg.info('doing audit of private key %r for remote user %r' % (key_id, remote_idurl))
        private_test_encrypted_sample = my_keys.encrypt(key_id, private_test_sample)
    json_payload = {
        'key_id': key_id,
        'audit': {
            'public_sample': '',
            'private_sample': base64.b64encode(private_test_encrypted_sample),
        },
    }
    raw_payload = serialization.DictToBytes(json_payload, values_to_text=True)
    block = encrypted.Block(
        BackupID=key_id,
        Data=raw_payload,
        SessionKey=key.NewSessionKey(session_key_type=key.SessionKeyType()),
        SessionKeyType=key.SessionKeyType(),
        # encrypt data using public key of recipient
        EncryptKey=lambda inp: recipient_id_obj.encrypt(inp),
    )
    encrypted_payload = block.Serialize()
    p2p_service.SendAuditKey(
        remote_idurl=remote_idurl,
        encrypted_payload=encrypted_payload,
        packet_id=key_id,
        timeout=timeout,
        callbacks={
            commands.Ack(): lambda response, info: _on_audit_private_key_response(response, info, key_id, untrusted_idurl, private_test_sample, result),
            commands.Fail(): lambda response, info: result.errback(Exception(response)),
            None: lambda pkt_out: result.errback(Exception('timeout')),  # timeout
        },
    )
    return result


#------------------------------------------------------------------------------


def on_key_received(newpacket, info, status, error_message):
    """
    Callback will be executed when I receive a new key from one remote user.
    """
    block = encrypted.Unserialize(newpacket.Payload)
    if block is None:
        lg.err('failed reading key info from %s' % newpacket.RemoteID)
        return False
    try:
        key_data = block.Data()
        key_json = serialization.BytesToDict(key_data, keys_to_text=True, values_to_text=True)
        # key_id = strng.to_text(key_json['key_id'])
        key_label = strng.to_text(key_json.get('label', ''))
        key_id, key_object = my_keys.read_key_info(key_json)
        if key_object.isSigned():
            if not my_keys.verify_key_info_signature(key_json):
                raise Exception('received key signature verification failed: %r' % key_json)
            # TODO: must also compare "signature_pubkey" with pub key of the creator of the key!
        if key_object.isPublic():
            # received key is a public key
            if my_keys.is_key_registered(key_id):
                # but we already have a key with that ID
                if my_keys.is_key_private(key_id):
                    # we should not overwrite existing private key
                    # TODO: check other scenarios
                    raise Exception('private key already registered with %r' % key_id)
                if my_keys.get_public_key_raw(key_id) != key_object.toPublicString():
                    my_keys.erase_key(key_id)
                    if not my_keys.register_key(key_id, key_object, label=key_label):
                        raise Exception('key register failed')
                    else:
                        lg.info('replaced existing key %s, is_public=%s' % (key_id, key_object.isPublic()))
                    # normally should not overwrite existing public key
                    # TODO: look more if need to add some extra checks
                    # for example need to be able to overwrite or erase remotely some keys to cleanup
                    # raise Exception('another public key already registered with %r and new key is not matching' % key_id)
                p2p_service.SendAck(newpacket)
                lg.warn('received existing public key: %s, skip' % key_id)
                return True
            if not my_keys.register_key(key_id, key_object, label=key_label):
                raise Exception('key register failed')
            else:
                lg.info('added new key %s, is_public=%s' % (key_id, key_object.isPublic()))
            p2p_service.SendAck(newpacket)
            if _Debug:
                lg.info('received and stored locally a new key %s, include_private=%s' % (key_id, key_json.get('include_private')))
            return True
        # received key is a private key
        if my_keys.is_key_registered(key_id):
            # check if we already have that key
            if my_keys.is_key_private(key_id):
                # we have already private key with same ID!!!
                if my_keys.get_private_key_raw(key_id) != key_object.toPrivateString():
                    # and this is a new private key : we should not overwrite!
                    raise Exception('private key already registered and it is not matching with received copy')
                # this is the same private key
                p2p_service.SendAck(newpacket)
                lg.warn('received again an exact copy of already existing private key: %s, skip' % key_id)
                return True
            # but we have a public key with same ID already
            # if my_keys.get_public_key_raw(key_id) != key_object.toPublicString():
            #     # and we should not overwrite existing public key as well
            #     raise Exception('another public key already registered with that ID and it is not matching with private key')
            lg.info('erasing public key %s' % key_id)
            my_keys.erase_key(key_id)
            if not my_keys.register_key(key_id, key_object, label=key_label):
                raise Exception('key register failed')
            lg.info('added new key %s, is_public=%s' % (key_id, key_object.isPublic()))
            p2p_service.SendAck(newpacket)
            return True
        # no private key with given ID was registered
        if not my_keys.register_key(key_id, key_object, label=key_label):
            raise Exception('key register failed')
        lg.info('added new key %s, is_public=%s' % (key_id, key_object.isPublic()))
        p2p_service.SendAck(newpacket)
        return True
    except Exception as exc:
        lg.exc()
        p2p_service.SendFail(newpacket, strng.to_text(exc))
    return False


def on_audit_key_received(newpacket, info, status, error_message):
    """
    Callback will be executed when remote user would like to check if I poses given key locally.
    """
    block = encrypted.Unserialize(newpacket.Payload)
    if block is None:
        lg.out(2, 'key_ring.on_audit_key_received ERROR reading data from %s' % newpacket.RemoteID)
        return False
    try:
        raw_payload = block.Data()
        json_payload = serialization.BytesToDict(raw_payload, keys_to_text=True, values_to_text=True)
        key_id = my_keys.latest_key_id(json_payload['key_id'])
        json_payload['audit']
        public_sample = base64.b64decode(json_payload['audit']['public_sample'])
        private_sample = base64.b64decode(json_payload['audit']['private_sample'])
    except Exception as exc:
        lg.exc()
        p2p_service.SendFail(newpacket, str(exc))
        return False
    if not my_keys.is_valid_key_id(key_id):
        p2p_service.SendFail(newpacket, 'invalid key id')
        return False
    if not my_keys.is_key_registered(key_id, include_master=True):
        p2p_service.SendFail(newpacket, 'key not registered')
        return False
    if public_sample:
        response_payload = base64.b64encode(my_keys.encrypt(key_id, public_sample))
        p2p_service.SendAck(newpacket, response_payload)
        if _Debug:
            lg.info('remote user %s requested audit of public key %s' % (newpacket.OwnerID, key_id))
        return True
    if private_sample:
        if not my_keys.is_key_private(key_id):
            p2p_service.SendFail(newpacket, 'private key not registered')
            return False
        response_payload = base64.b64encode(my_keys.decrypt(key_id, private_sample))
        p2p_service.SendAck(newpacket, response_payload)
        if _Debug:
            lg.info('remote user %s requested audit of private key %s' % (newpacket.OwnerID, key_id))
        return True
    p2p_service.SendFail(newpacket, 'wrong audit request')
    return False


#------------------------------------------------------------------------------


def do_backup_key(key_id, keys_folder=None):
    """
    Send given key to my suppliers to store it remotely.
    This will make a regular backup copy of that key file - encrypted with my master key.
    """
    key_id = my_keys.latest_key_id(key_id)
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.do_backup_key     key_id=%r' % key_id)
    if key_id == my_id.getGlobalID(key_alias='master') or key_id == 'master':
        lg.err('master key must never leave local host')
        return fail(Exception('master key must never leave local host'))
    if not my_keys.is_key_registered(key_id):
        lg.err('unknown key: "%s"' % key_id)
        return fail(Exception('unknown key: "%s"' % key_id))
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if my_keys.is_key_private(key_id):
        local_key_filepath = os.path.join(keys_folder, '%s.private' % key_id)
        remote_path_for_key = '.keys/%s.private' % key_id
    else:
        local_key_filepath = os.path.join(keys_folder, '%s.public' % key_id)
        remote_path_for_key = '.keys/%s.public' % key_id
    global_key_path = global_id.MakeGlobalID(key_alias='master', customer=my_id.getGlobalID(), path=remote_path_for_key)
    res = api.file_exists(global_key_path)
    if res['status'] == 'OK' and res['result'] and res['result'].get('exist'):
        from bitdust.storage import backup_control
        lg.warn('key %s already exists in catalog' % global_key_path)
        global_key_path_id = res['result'].get('path_id')
        if global_key_path_id and backup_control.IsPathInProcess(global_key_path_id):
            lg.warn('skip, another backup for key already started: %s' % global_key_path_id)
            backup_id_list = backup_control.FindRunningBackup(global_key_path_id)
            if backup_id_list:
                backup_id = backup_id_list[0]
                backup_job = backup_control.GetRunningBackupObject(backup_id)
                if backup_job:
                    backup_result = Deferred()
                    backup_job.resultDefer.addCallback(lambda resp: backup_result.callback(True) if resp == 'done' else backup_result.errback(Exception('failed to upload key "%s", task was not started: %r' % (global_key_path, resp))))
                    backup_job.resultDefer.addErrback(lg.errback, debug=_Debug, debug_level=_DebugLevel, method='key_ring.do_backup_key')
                    backup_job.resultDefer.addErrback(backup_result.errback)
                    if _Debug:
                        lg.args(_DebugLevel, backup_id=backup_id, global_key_path_id=global_key_path_id)
                    return backup_result
                else:
                    lg.warn('did not found running backup job: %r' % backup_id)
            else:
                lg.warn('did not found running backup id for path: %r' % global_key_path_id)
    else:
        res = api.file_create(global_key_path)
        if res['status'] != 'OK':
            lg.err('failed to create path "%s" in the catalog: %r' % (global_key_path, res))
            return fail(Exception('failed to create path "%s" in the catalog: %r' % (global_key_path, res)))
    res = api.file_upload_start(
        local_path=local_key_filepath,
        remote_path=global_key_path,
        wait_result=True,
    )
    backup_result = Deferred()

    # TODO: put that code bellow into api.file_upload_start() method with additional parameter

    def _job_done(result):
        if _Debug:
            lg.args(_DebugLevel, key_id=key_id, result=result)
        if result == 'done':
            backup_result.callback(True)
        else:
            backup_result.errback(Exception('failed to upload key "%s", backup is %r' % (key_id, result)))
        return None

    def _task_started(resp):
        if _Debug:
            lg.args(_DebugLevel, key_id=key_id, upload_response_status=resp['status'])
        if resp['status'] != 'OK':
            backup_result.errback(Exception('failed to upload key "%s", task was not started: %r' % (global_key_path, resp)))
            return None
        backupObj = backup_control.jobs().get(resp['version'])
        if not backupObj:
            backup_result.errback(Exception('failed to upload key "%s", task %r failed to start' % (global_key_path, resp['version'])))
            return None
        backupObj.resultDefer.addCallback(_job_done)
        backupObj.resultDefer.addErrback(backup_result.errback)
        return None

    if not isinstance(res, Deferred):
        res_defer = Deferred()
        res_defer.callback(res)
        res = res_defer
    res.addCallback(_task_started)
    res.addErrback(backup_result.errback)
    return backup_result


def do_restore_key(key_id, is_private, keys_folder=None, wait_result=False):
    """
    Restore given key from my suppliers if I do not have it locally.
    """
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.do_restore_key     key_id=%r    is_private=%r' % (key_id, is_private))
    key_id = my_keys.latest_key_id(key_id)
    if my_keys.is_key_registered(key_id):
        lg.err('local key already exist: "%s"' % key_id)
        if wait_result:
            return fail(Exception('local key already exist: "%s"' % key_id))
        return False
    if not keys_folder:
        keys_folder = settings.KeyStoreDir()
    if is_private:
        remote_path_for_key = '.keys/%s.private' % key_id
    else:
        remote_path_for_key = '.keys/%s.public' % key_id
    global_key_path = global_id.MakeGlobalID(key_alias='master', customer=my_id.getGlobalID(), path=remote_path_for_key)
    ret = api.file_download_start(
        remote_path=global_key_path,
        destination_path=keys_folder,
        wait_result=True,
    )
    if not isinstance(ret, Deferred):
        lg.err('failed to download key "%s": %s' % (key_id, ret))
        if wait_result:
            return fail(Exception('failed to download key "%s": %s' % (key_id, ret)))
        return False

    result = Deferred()

    def _on_result(res):
        if not isinstance(res, dict):
            lg.err('failed to download key "%s": %s' % (key_id, res))
            if wait_result:
                result.errback(Exception('failed to download key "%s": %s' % (key_id, res)))
            return None
        if res['status'] != 'OK':
            lg.err('failed to download key "%s": %r' % (key_id, res))
            if wait_result:
                result.errback(Exception('failed to download key "%s": %r' % (key_id, res)))
            return None
        if not my_keys.load_key(key_id, keys_folder):
            lg.err('failed to read key "%s" from local folder "%s"' % (key_id, keys_folder))
            if wait_result:
                result.errback(Exception('failed to read key "%s" from local folder "%s"' % (key_id, keys_folder)))
            return None
        if _Debug:
            lg.out(_DebugLevel, 'key_ring.do_restore_key._on_result key_id=%s  is_private=%r : %r' % (key_id, is_private, res))
        if wait_result:
            result.callback(res)
        return None

    ret.addBoth(_on_result)

    if not wait_result:
        return True
    return result


def do_delete_key(key_id, is_private):
    """
    Remove given key from my suppliers nodes.
    """
    if is_private:
        remote_path_for_key = '.keys/%s.private' % key_id
    else:
        remote_path_for_key = '.keys/%s.public' % key_id
    global_key_path = global_id.MakeGlobalID(key_alias='master', customer=my_id.getGlobalID(), path=remote_path_for_key)
    res = api.file_delete(global_key_path)
    if res['status'] != 'OK':
        lg.err('failed to delete key "%s": %r' % (global_key_path, res))
        return False
    if _Debug:
        lg.out(_DebugLevel, 'key_ring.do_delete_key key_id=%s  is_private=%r : %r' % (key_id, is_private, res))
    return True


#------------------------------------------------------------------------------


def on_files_received(newpacket, info):
    list_files_global_id = global_id.ParseGlobalID(newpacket.PacketID)
    if not list_files_global_id['idurl']:
        lg.warn('invalid PacketID: %s' % newpacket.PacketID)
        return False
    incoming_key_id = list_files_global_id['key_id']
    if not my_keys.is_valid_key_id(incoming_key_id):
        lg.warn('ignore, invalid key id in packet %s' % newpacket)
        return False
    if list_files_global_id['key_alias'] == 'master':
        if _Debug:
            lg.dbg(_DebugLevel, 'ignore %s packet which seems to came from my own supplier' % newpacket)
        # only process list Files() from other customers who granted me access to their files
        return False
    if not my_keys.is_key_private(incoming_key_id):
        lg.warn('private key is not registered : %s' % incoming_key_id)
        p2p_service.SendFail(newpacket, 'private key is not registered')
        return False
    if list_files_global_id['key_alias'].startswith('share_'):
        from bitdust.access import shared_access_coordinator
        return shared_access_coordinator.on_list_files_verified(newpacket, list_files_global_id)
    # for other shared files that are not controlled by shared_access_coordinator(): message archive keys are starting with "group_"
    try:
        block = encrypted.Unserialize(newpacket.Payload, decrypt_key=incoming_key_id)
    except:
        lg.exc(newpacket.Payload)
        return False
    if block is None:
        lg.warn('failed reading data from %s' % newpacket.RemoteID)
        return False
    try:
        raw_files = block.Data()
    except:
        lg.exc()
        return False
    from bitdust.storage import index_synchronizer
    from bitdust.storage import backup_control
    from bitdust.storage import backup_fs
    # otherwise this must be an external supplier sending us a files he stores for trusted customer
    external_supplier_idurl = block.CreatorID
    try:
        supplier_raw_list_files = backup_control.UnpackListFiles(raw_files, settings.ListFilesFormat())
    except:
        lg.exc()
        return False
    # need to detect supplier position from the list of packets
    # and place that supplier on the correct position in contactsdb
    supplier_pos = backup_matrix.DetectSupplierPosition(supplier_raw_list_files)
    trusted_customer_idurl = list_files_global_id['idurl']
    known_supplier_pos = contactsdb.supplier_position(external_supplier_idurl, trusted_customer_idurl)
    if known_supplier_pos < 0:
        lg.warn('received %r from an unknown node %r which is not a supplier of %r' % (newpacket, external_supplier_idurl, trusted_customer_idurl))
        return False
    if _Debug:
        lg.args(_DebugLevel, sz=len(supplier_raw_list_files), s=external_supplier_idurl, pos=supplier_pos, known_pos=known_supplier_pos, pid=newpacket.PacketID)
    if supplier_pos >= 0:
        if known_supplier_pos != supplier_pos:
            lg.err('known external supplier %r position %d is not matching with received list files position %d for customer %s' % (external_supplier_idurl, known_supplier_pos, supplier_pos, trusted_customer_idurl))
            return False
    else:
        lg.warn('not possible to detect external supplier position for customer %s from received list files, known position is %s' % (trusted_customer_idurl, known_supplier_pos))
        supplier_pos = known_supplier_pos
    is_in_sync = index_synchronizer.is_synchronized() and backup_fs.revision() > 0
    remote_files_changed, _, _, _ = backup_matrix.process_raw_list_files(
        supplier_num=supplier_pos,
        list_files_text_body=supplier_raw_list_files,
        customer_idurl=trusted_customer_idurl,
        is_in_sync=is_in_sync,
    )
    if remote_files_changed:
        backup_matrix.SaveLatestRawListFiles(
            supplier_idurl=external_supplier_idurl,
            raw_data=supplier_raw_list_files,
            customer_idurl=trusted_customer_idurl,
        )
    # finally sending Ack() packet back
    p2p_service.SendAck(newpacket)
    if remote_files_changed:
        lg.info('received updated list of files from external supplier %s for customer %s' % (external_supplier_idurl, trusted_customer_idurl))
    return True



#!/usr/bin/env python
# getsizeof.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (getsizeof.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
# taken from:
# http://code.activestate.com/recipes/577504/

from __future__ import absolute_import
from __future__ import print_function
from sys import getsizeof
from itertools import chain
from collections import deque
from six.moves import map
try:
    import reprlib
except:
    pass

#------------------------------------------------------------------------------


def total_size(o, handlers={}, verbose=False):
    """
    Returns the approximate memory footprint an object and all of its contents.

    Automatically finds the contents of the following builtin containers and
    their subclasses:  tuple, list, deque, dict, set and frozenset.
    To search other containers, add handlers to iterate over their contents:

        handlers = {SomeContainerClass: iter,
                    OtherContainerClass: OtherContainerClass.get_elements}
    """
    dict_handler = lambda d: chain.from_iterable(list(d.items()))
    all_handlers = {
        tuple: iter,
        list: iter,
        deque: iter,
        dict: dict_handler,
        set: iter,
        frozenset: iter,
    }
    all_handlers.update(handlers)  # user handlers take precedence
    seen = set()  # track which object id's have already been seen
    default_size = getsizeof(0)  # estimate sizeof object without __sizeof__

    def sizeof(o):
        if id(o) in seen:  # do not double count the same object
            return 0
        seen.add(id(o))
        s = getsizeof(o, default_size)

        if verbose:
            print((s, type(o), reprlib.repr(o)))  # , file=stderr)

        for typ, handler in all_handlers.items():
            if isinstance(o, typ):
                s += sum(map(sizeof, handler(o)))
                break
        return s

    return sizeof(o)


##### Example call #####

if __name__ == '__main__':
    d = dict(a=1, b=2, c=3, d=[4, 5, 6, 7], e='a string of chars')
    print((total_size(d, verbose=True)))



#!/usr/bin/env python
# pptransport.py
#
# Parallel Python Software: http://www.parallelpython.com
# Copyright (c) 2005-2009, Vitalii Vanovschi
# All rights reserved.
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#    * Redistributions of source code must retain the above copyright notice,
#      this list of conditions and the following disclaimer.
#    * Redistributions in binary form must reproduce the above copyright
#      notice, this list of conditions and the following disclaimer in the
#      documentation and/or other materials provided with the distribution.
#    * Neither the name of the author nor the names of its contributors
#      may be used to endorse or promote products derived from this software
#      without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT OWNER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF
# THE POSSIBILITY OF SUCH DAMAGE.
"""
Parallel Python Software, PP Transport.

http://www.parallelpython.com - updates, documentation, examples and support
forums
"""
from __future__ import absolute_import
import six

copyright = 'Copyright (c) 2005-2009 Vitalii Vanovschi. All rights reserved'
version = '1.5.7'

import struct
import socket
import logging

# compartibility with Python 2.6
try:
    import hashlib
    sha_new = hashlib.sha1
    md5_new = hashlib.md5
except ImportError:
    import sha
    import md5
    sha_new = sha.new
    md5_new = md5.new


class Transport(object):
    def send(self, msg):
        raise NotImplemented("abstact function 'send' must be implemented "
                             'in a subclass', )

    def receive(self, preprocess=None):
        raise NotImplemented("abstact function 'receive' must be implemented "
                             'in a subclass', )

    def authenticate(self, secret):
        remote_version = self.receive()
        if version != remote_version:
            logging.error('PP version mismatch (local: pp-%s, remote: pp-%s)' % (version, remote_version))
            logging.error('Please install the same version of PP on all nodes')
            return False
        srandom = self.receive()
        answer = sha_new(srandom + secret).hexdigest()
        self.send(answer)
        response = self.receive()
        if response == 'OK':
            return True
        else:
            return False

    def close(self):
        pass

    def _connect(self, host, port):
        pass


class CTransport(Transport):
    """
    Cached transport.
    """
    rcache = {}

    def hash(self, msg):
        return md5_new(msg).hexdigest()

    def csend(self, msg):
        hash1 = self.hash(msg)
        if hash1 in self.scache:
            self.send(b'H' + hash1)
        else:
            self.send(b'N' + msg)
            self.scache[hash1] = True

    def creceive(self, preprocess=None):
        msg = self.receive()
        if msg[0] == b'H':
            hash1 = msg[1:]
        else:
            msg = msg[1:]
            hash1 = self.hash(msg)
            self.rcache[hash1] = [r for r in map(preprocess or (lambda m: m), (msg, ))][0]
        return self.rcache[hash1]


class PipeTransport(Transport):
    def __init__(self, r, w):
        self.scache = {}
        self.exiting = False
        if hasattr(r, 'read') and hasattr(w, 'write'):
            self.r = r
            self.w = w
        else:
            raise TypeError('Both arguments of PipeTransport constructor '
                            'must be file objects', )

    def send(self, msg):
        if not isinstance(msg, six.binary_type):
            msg = msg.encode('latin1')
        msg_size = struct.pack('!Q', len(msg))
        if not isinstance(msg_size, six.binary_type):
            msg_size = msg_size.encode('latin1')
        self.w.write(msg_size)
        self.w.flush()
        self.w.write(msg)
        self.w.flush()

    def receive(self, preprocess=None):
        first_bytes = struct.calcsize('!Q')
        size_packed = self.r.read(first_bytes)
        if not isinstance(size_packed, six.binary_type):
            size_packed = size_packed.encode('latin1')
        msg_len = struct.unpack('!Q', size_packed)[0]
        msg = self.r.read(msg_len)
        if isinstance(msg, six.binary_type):
            msg = msg.decode('latin1')
        return [r for r in map(preprocess or (lambda m: m), (msg, ))][0]

    def close(self):
        self.w.close()
        self.r.close()


class SocketTransport(Transport):
    def __init__(self, socket1=None):
        if socket1:
            self.socket = socket1
        else:
            self.socket = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
        self.scache = {}

    def send(self, data):
        size = struct.pack('!Q', len(data))
        t_size = struct.calcsize('!Q')
        s_size = 0
        while s_size < t_size:
            p_size = self.socket.send(size[s_size:])
            if p_size == 0:
                raise RuntimeError('Socket connection is broken')
            s_size += p_size

        t_size = len(data)
        s_size = 0
        while s_size < t_size:
            p_size = self.socket.send(data[s_size:])
            if p_size == 0:
                raise RuntimeError('Socket connection is broken')
            s_size += p_size

    def receive(self, preprocess=None):
        e_size = struct.calcsize('!Q')
        r_size = 0
        data = b''
        while r_size < e_size:
            msg = self.socket.recv(e_size - r_size)
            if not msg:
                raise RuntimeError('Socket connection is broken')
            r_size += len(msg)
            data += msg
        e_size = struct.unpack('!Q', data)[0]

        r_size = 0
        data = b''
        while r_size < e_size:
            msg = self.socket.recv(e_size - r_size)
            if not msg:
                raise RuntimeError('Socket connection is broken')
            r_size += len(msg)
            data += msg
        return data

    def close(self):
        self.socket.close()

    def _connect(self, host, port):
        self.socket.connect((host, port))


class CPipeTransport(PipeTransport, CTransport):
    pass


class CSocketTransport(SocketTransport, CTransport):
    pass


# Parallel Python Software: http://www.parallelpython.com



#!/usr/bin/env python
# setup.py
#
# Copyright (C) 2008 Veselin Penev, http://bitdust.io
#
# This file (setup.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
from __future__ import absolute_import
from setuptools import setup

setup(
    name='bitdust',
    version='0.1.13.247',
    author='Veselin Penev, BitDust',
    author_email='bitdust.io@gmail.com',
    maintainer='Veselin Penev, BitDust',
    maintainer_email='veselin@bitdust.io',
    url='https://bitdust.io',
    description='p2p secure distributed storage and communication engine',
    long_description=open('README.txt', 'r').read(),
    long_description_content_type='text/markdown',
    download_url='https://bitdust.io',
    license='Copyright (C) 2008 Veselin Penev, https://bitdust.io',

    keywords='''p2p, peer to peer, backup, restore, storage, data, recover,
                distributed, online, python, twisted, messaging, websocket,
                encryption, crypto, protection''',

    classifiers=[
        'Development Status :: 3 - Alpha',
        'Environment :: Console',
        'Environment :: Web Environment',
        'Framework :: Twisted',
        'Intended Audience :: Customer Service',
        'Intended Audience :: Developers',
        'Intended Audience :: End Users/Desktop',
        'Intended Audience :: Information Technology',
        'Intended Audience :: System Administrators',
        'License :: Other/Proprietary License',
        'Natural Language :: English',
        'Operating System :: Microsoft :: Windows',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
        'Topic :: Security',
        'Topic :: Security :: Cryptography',
        'Topic :: System :: Archiving :: Backup',
        'Topic :: System :: Distributed Computing',
        'Topic :: Utilities',
    ],

    packages=[
        'bitdust',
        'bitdust.access',
        'bitdust.automats',
        'bitdust.blockchain',
        'bitdust.broadcast',
        'bitdust.chat',
        'bitdust.coins',
        'bitdust.contacts',
        'bitdust.crypt',
        'bitdust.currency',
        'bitdust.customer',
        'bitdust.dht',
        'bitdust.interface',
        'bitdust.lib',
        'bitdust.logs',
        'bitdust.main',
        'bitdust.p2p',
        'bitdust.raid',
        'bitdust.services',
        'bitdust.storage',
        'bitdust.stream',
        'bitdust.stun',
        'bitdust.supplier',
        'bitdust.system',
        'bitdust.transport',
        'bitdust.transport.http',
        'bitdust.transport.proxy',
        'bitdust.transport.udp',
        'bitdust.transport.tcp',
        'bitdust.updates',
        'bitdust.userid',
        'bitdust_forks',
        'bitdust_forks.Bismuth',
        'bitdust_forks.CodernityDB',
        'bitdust_forks.CodernityDB3',
        'bitdust_forks.entangled',
        'bitdust_forks.entangled.kademlia',
        'bitdust_forks.parallelp',
        'bitdust_forks.parallelp.pp',
        'bitdust_forks.txrestapi',
        'bitdust_forks.txrestapi.txrestapi',
        'bitdust_forks.websocket',
    ],

    package_data={
        'bitdust': [
            '*.txt',
            '*.md',
            '*.json',
        ],
        'bitdust_forks': [
            'bitdust_forks/Bismuth/LICENSE',
            'bitdust_forks/Bismuth/README.md',
            'bitdust_forks/CodernityDB/README',
            'bitdust_forks/entangled/AUTHORS',
            'bitdust_forks/entangled/COPYING',
            'bitdust_forks/entangled/README',
            'bitdust_forks/parallelp/README',
            'bitdust_forks/parallelp/pp/AUTHORS',
            'bitdust_forks/parallelp/pp/CHANGELOG',
            'bitdust_forks/parallelp/pp/COPYING',
            'bitdust_forks/parallelp/pp/PKG-INFO',
            'bitdust_forks/parallelp/pp/README',
            'bitdust_forks/txrestapi/LICENSE',
            'bitdust_forks/txrestapi/README.rst',
            'bitdust_forks/websocket/LICENSE',
        ],
    },

    install_requires=[
        'Twisted==20.3.0',
        'zope.interface',
        'cryptography',
        'pycryptodomex',
        'service_identity',
        'pyparsing',
        'appdirs',
        'psutil',
        'cffi',
        'six',
        'virtualenv',
    ],

    scripts=[
        'scripts/bitdust',
        'scripts/bitdust_worker',
    ],

)



#!/usr/bin/python
# service_message_broker.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_message_broker.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_message_broker
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return MessageBrokerService()


class MessageBrokerService(LocalService):

    service_name = 'service_message_broker'
    config_path = 'services/message-broker/enabled'
    data_dir_required = True

    last_time_keys_synchronized = None

    def dependent_on(self):
        return [
            'service_private_messages',
            'service_data_disintegration',
        ]

    def installed(self):
        return False

    def attached_dht_layers(self):
        from bitdust.dht import dht_records
        return [
            dht_records.LAYER_MESSAGE_BROKERS,
        ]

    def start(self):
        from bitdust.main import events
        from bitdust.stream import message_peddler
        message_peddler.A('start')
        # self._do_connect_message_brokers_dht_layer()
        # events.add_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        events.add_subscriber(self._on_my_identity_url_changed, 'my-identity-url-changed')
        return True

    def stop(self):
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.main import events
        from bitdust.stream import message_peddler
        events.remove_subscriber(self._on_my_identity_url_changed, 'my-identity-url-changed')
        # events.remove_subscriber(self._on_dht_layer_connected, 'dht-layer-connected')
        # dht_service.suspend(layer_id=dht_records.LAYER_MESSAGE_BROKERS)
        message_peddler.A('stop')
        return True

    def health_check(self):
        return True

    def request(self, json_payload, newpacket, info):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.p2p import p2p_service
        from bitdust.stream import message_peddler
        from bitdust.userid import id_url
        result = Deferred()
        try:
            action = json_payload['action']
        except:
            lg.exc()
            return p2p_service.SendFail(newpacket, 'invalid payload')
        # TODO: validate signature and the key
        if action == 'queue-connect' or action == 'queue-connect-follow':
            try:
                queue_id = json_payload.get('queue_id')
                consumer_id = json_payload.get('consumer_id')
                producer_id = json_payload.get('producer_id')
                group_key = json_payload.get('group_key')
                position = int(json_payload.get('position', -1))
                archive_folder_path = json_payload.get('archive_folder_path', None)
                last_sequence_id = json_payload.get('last_sequence_id', -1)
                known_brokers = json_payload.get('known_brokers', {}) or {}
                known_brokers = {int(k): id_url.field(v) for k, v in known_brokers.items()}
            except:
                lg.warn('wrong payload: %r' % json_payload)
                return p2p_service.SendFail(newpacket, 'wrong payload')
            message_peddler.A(
                event='connect' if action == 'queue-connect' else 'follow',
                queue_id=queue_id,
                consumer_id=consumer_id,
                producer_id=producer_id,
                group_key=group_key,
                position=position,
                archive_folder_path=archive_folder_path,
                last_sequence_id=last_sequence_id,
                known_brokers=known_brokers,
                request_packet=newpacket,
                result_defer=result,
            )
        elif action == 'broker-verify':
            try:
                customer_id = json_payload['customer_id']
                broker_id = json_payload['broker_id']
                position = json_payload['position']
                known_streams = json_payload['streams']
                known_brokers = {int(k): id_url.field(v) for k, v in json_payload['known_brokers'].items()}
            except:
                lg.warn('wrong payload: %r' % json_payload)
                return p2p_service.SendFail(newpacket, 'wrong payload')
            message_peddler.A(
                event='broker-reconnect',
                customer_id=customer_id,
                broker_id=broker_id,
                position=position,
                known_streams=known_streams,
                known_brokers=known_brokers,
                request_packet=newpacket,
                result_defer=result,
            )
        else:
            lg.warn('wrong action request' % newpacket.Payload)
            return p2p_service.SendFail(newpacket, 'wrong action request')
        return result

    def cancel(self, json_payload, newpacket, info):
        from twisted.internet.defer import Deferred
        from bitdust.logs import lg
        from bitdust.p2p import p2p_service
        from bitdust.stream import message_peddler
        try:
            action = json_payload['action']
            queue_id = json_payload.get('queue_id', None)
            consumer_id = json_payload['consumer_id']
            producer_id = json_payload['producer_id']
            group_key = json_payload['group_key']
        except:
            lg.warn('wrong payload: %r' % json_payload)
            return p2p_service.SendFail(newpacket, 'wrong payload')
        # TODO: validate signature and the key
        result = Deferred()
        if action == 'queue-disconnect':
            message_peddler.A(
                event='disconnect',
                queue_id=queue_id,
                consumer_id=consumer_id,
                producer_id=producer_id,
                group_key=group_key,
                request_packet=newpacket,
                result_defer=result,
            )
        else:
            lg.warn('wrong action request' % newpacket.Payload)
            return p2p_service.SendFail(newpacket, 'wrong action request')
        return result

    def _do_connect_message_brokers_dht_layer(self):
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.dht import known_nodes
        known_seeds = known_nodes.nodes()
        d = dht_service.open_layer(
            layer_id=dht_records.LAYER_MESSAGE_BROKERS,
            seed_nodes=known_seeds,
            connect_now=True,
            attach=True,
        )
        d.addCallback(self._on_message_brokers_dht_layer_connected)
        d.addErrback(lambda *args: lg.err(str(args)))

    def _on_message_brokers_dht_layer_connected(self, ok):
        from bitdust.logs import lg
        from bitdust.dht import dht_service
        from bitdust.dht import dht_records
        from bitdust.userid import my_id
        lg.info('connected to DHT layer for message brokers: %r' % ok)
        if my_id.getIDURL():
            dht_service.set_node_data('idurl', my_id.getIDURL().to_text(), layer_id=dht_records.LAYER_MESSAGE_BROKERS)
        return ok

    def _on_dht_layer_connected(self, evt):
        from bitdust.dht import dht_records
        if evt.data['layer_id'] == 0:
            self._do_connect_message_brokers_dht_layer()
        elif evt.data['layer_id'] == dht_records.LAYER_MESSAGE_BROKERS:
            self._on_message_brokers_dht_layer_connected(True)

    def _on_my_identity_url_changed(self, evt):
        from bitdust.stream import message_peddler
        message_peddler.A('stop')
        message_peddler.close_all_streams()
        message_peddler.check_rotate_queues()



#WARNING: this file uses the old wallet structure, there is no need to update it at the moment

import hashlib
import sqlite3
import os
import sys
import time
import base64

from Cryptodome.PublicKey import RSA
from Cryptodome.Signature import PKCS1_v1_5
from Cryptodome.Hash import SHA

DIFFICULTY = 10

if os.path.isfile('privkey.der'):
    print('privkey.der found')
elif os.path.isfile('privkey_encrypted.der'):
    print('privkey_encrypted.der found')

else:
    # generate key pair and an address
    key = RSA.generate(4096)
    public_key = key.publickey()

    private_key_readable = str(key.exportKey().decode('utf-8'))
    public_key_readable = str(key.publickey().exportKey().decode('utf-8'))
    address = hashlib.sha224(public_key_readable.encode('utf-8')).hexdigest()  # hashed public key
    # generate key pair and an address

    print('Your address: {}'.format(address))
    print('Your private key:\n {}'.format(private_key_readable))
    print('Your public key:\n {}'.format(public_key_readable))

    with open('privkey.der', 'a') as f:
        f.write(str(private_key_readable))

    with open('pubkey.der', 'a') as f:
        f.write(str(public_key_readable))

    with open('address.txt', 'a') as f:
        f.write('{}\n'.format(address))

# import keys
key = RSA.importKey(open('privkey.der').read())
public_key = key.publickey()
private_key_readable = str(key.exportKey().decode('utf-8'))
public_key_readable = str(key.publickey().exportKey().decode('utf-8'))
address = hashlib.sha224(public_key_readable.encode('utf-8')).hexdigest()

print('Your address: {}'.format(address))
print('Your private key:\n {}'.format(private_key_readable))
print('Your public key:\n {}'.format(public_key_readable))
public_key_b64encoded = base64.b64encode(public_key_readable.encode('utf-8'))
# import keys

timestamp = str(time.time())
print('Timestamp: {}'.format(timestamp))
transaction = (timestamp, 'genesis', address, str(float(100000000)), 'genesis')
h = SHA.new(str(transaction).encode('utf-8'))
signer = PKCS1_v1_5.new(key)
signature = signer.sign(h)
signature_enc = base64.b64encode(signature)
print('Encoded Signature: {}'.format(signature_enc))
block_hash = hashlib.sha224(str((timestamp, transaction)).encode('utf-8')).hexdigest()  # first hash is simplified
print('Transaction Hash: {}'.format(block_hash))

if os.path.isfile('static/ledger.db'):
    print('You are beyond genesis')
else:
    # transaction processing
    cursor = None
    mem_cur = None
    try:
        conn = sqlite3.connect('static/ledger.db')
        cursor = conn.cursor()

        cursor.execute('CREATE TABLE IF NOT EXISTS "misc" ("block_height" INTEGER, "difficulty" TEXT)')
        cursor.execute(
            'CREATE TABLE IF NOT EXISTS "transactions" ("block_height" INTEGER, "timestamp" NUMERIC, "address" TEXT, "recipient" TEXT, "amount" NUMERIC, "signature" TEXT, "public_key" TEXT, "block_hash" TEXT, "fee" NUMERIC, "reward" NUMERIC, "operation" TEXT, "openfield" TEXT)',
        )
        cursor.execute('CREATE INDEX "Timestamp Index" ON "transactions" ("timestamp")')
        cursor.execute('CREATE INDEX "Signature Index" ON "transactions" ("signature")')
        cursor.execute('CREATE INDEX "Reward Index" ON "transactions" ("reward")')
        cursor.execute('CREATE INDEX "Recipient Index" ON "transactions" ("recipient")')
        cursor.execute('CREATE INDEX "Openfield Index" ON "transactions" ("openfield")')
        cursor.execute('CREATE INDEX "Fee Index" ON "transactions" ("fee")')
        cursor.execute('CREATE INDEX "Block Height Index" ON "transactions" ("block_height")')
        cursor.execute('CREATE INDEX "Block Hash Index" ON "transactions" ("block_hash")')
        cursor.execute('CREATE INDEX "Amount Index" ON "transactions" ("amount")')
        cursor.execute('CREATE INDEX "Address Index" ON "transactions" ("address")')
        cursor.execute('CREATE INDEX "Operation Index" ON "transactions" ("operation")')
        cursor.execute('CREATE INDEX TXID4_Index ON transactions(substr(signature,1,4))')
        cursor.execute('CREATE INDEX "Misc Block Height Index" on misc(block_height)')
        # cursor.execute("INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)", (
        #     '1', '1574712979.530888', 'genesis', 'bc33dfda25a44f11ad41875e8bac85a7e876436f1d308130e0bc00f6', '0',
        #     'cacdDgpCAvSa04B9De522kvGmGKL1HG85XInoEwlSEOCGLPTI+lMjdc6VHPFakmyPFT5fosnWhdHJx8rAJh5UUmmZgNd6jFf1N5hom/PAJLGVyDmXXbcH1YzNvBXUJdwBV/M3YtK+t5hWr9D10tw3ahywNwQddAO4IJj6jeFuc9ApJUhRNpMy7PIgyaisMIJPylBGvSU4Wbyxx2DDSXTJ2kLHn4vJgRtpcSCTHFlCtVwVF0GpeWkwilKA+AgcFtLRvB4ead1w68diX+SocLXNueccAzlDnYL5h5C/QaP9AG8cBquHh4VLruBoXcRzKXjONKMIqhJKAgOdjo0Tmtrp7XH8PGYfmHAVChXHyE3lHm3qadAW5AFWNQigNHd5imlTl1Tz8wjBI/W5epag7WCr/s+3HkwdNY7eLiHCvp1lrWIuFqnLDVASIjIjMKffv+j3VKKXVh7wLeM4VuOqs6EQnmUpyXqQvJPz8ZJvaTWVzyVnEXteqfzG5UbSYhjZV10pDUlhxVfx3IfvdFzm8viZPTaI+8HNy8VWED0tLQRwlz/147aC2rtz0CBgO7tM3NdD9SGJzQ3l0iB055i32YREH+DpE9QpM9mDv+tAStNjE/ZuHsNehtOC3CED15N410aHtXyXC0VvbxL6bKH7FKeJEq+HcAHxVq5+JxJzsIE5pE=',
        #     'LS0tLS1CRUdJTiBQVUJMSUMgS0VZLS0tLS0KTUlJQ0lqQU5CZ2txaGtpRzl3MEJBUUVGQUFPQ0FnOEFNSUlDQ2dLQ0FnRUF2OUpSSmJ0REJsMTc3OTBxWjB1TwpTRW0rVUZFOXU3SENIb2JlblFmc0RNWEFyZ3J1dTFmdlM0VUh0RW55VFI0dnJsNGpybUxJNmVhaCtDTjZHSno0CnRja1ZTV2FQOTNWVm5iNkpPTStTdzdTckd6MHlOR1E5KzVRMXI2NXZNVjZKWWQrK3VTNjJPUDdkZHN1MzJGMTcKK2d0UEV2OUIrNUZ6bm9ya25BRzh2SitEVmFHSTh1RGt3bzFZSXlHL0VNVCtXaVZJcGI4UWZFNzlxK2NKdHF6agpkNklEYWFhdjdSWEpxZU04Q2ZTR05rYWFvSEZ0YStBZFV3Q0syVHdtOW5JZjBqRWxwVENHcnNjMzRLYVlMRGh0CmVWbkZSRWVsbGhyR2FiMmgwR3M2azdPbmJ5TXoxYkFhb0dGNk9acG45MEkwYmVGRmZuN3dnWXZJWUMwanNIcVEKZGowZlhhUGo1bGRETGR4bHJmWVBwcTVacXRyVEcyeU1FTmhYRzZYbFEyOENZS0psRGR3SmRFMHRTWm1pc1lnVwoxVnlEYkxtUjUzQ3JSR1dTTENrY05BSWlSRHYrM2ZSRXlWV0d3MmM3bEEwVHFueFJyN01rZFhiTjZiRDJSTVhECk00ZFBHWTZzZllMVU9OZGFUYS82ZkduR0NqTXRBNVZGTXpuUEhBcmdjSmpJSzZDeGJ6aTYxVnA0QzFJRzF4bk8KYkJIM1BtSUE3Qmh0ZEtWUmxucWZ2M1p6d1g0RFVEWEdLTm5qNVlKQ2pvUm9pUFN4ZW85ajM2OFBkRnU5OHVLcgpRTGpqZFo3b1hFMTVxVW5TeWIrWjdXQjlHNnBQT3dtRjQ0T0dka2lkQjRCS1pzb2Fzb09oUVBlMUhpVTVLdGh3CmlNQ0UzZnJiUDlJSTBtMWNWYWZUbHlVQ0F3RUFBUT09Ci0tLS0tRU5EIFBVQkxJQyBLRVktLS0tLQ==',
        #     'c5adf0949d1e678de1e9a8708045a474eb8022f8029bf01e7e975e5a', 0, 1, 1, 'genesis',
        # ))
        cursor.execute('INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)', ('1', timestamp, 'genesis', address, '0', signature_enc.decode('utf-8'), public_key_b64encoded, block_hash, 0, 1, 1, 'genesis'))  # Insert a row of data
        cursor.execute('INSERT INTO misc (difficulty, block_height) VALUES ({},1)'.format(DIFFICULTY))
        conn.commit()  # Save (commit) the changes

        mempool = sqlite3.connect('mempool.db')
        mem_cur = mempool.cursor()
        mem_cur.execute('CREATE TABLE transactions (timestamp, address, recipient, amount, signature, public_key, operation, openfield)')
        mempool.commit()
        conn.close()
        mempool.close()
        cursor = None
        mem_cur = None

        print('Genesis created.')
    except sqlite3.Error as e:
        print('Error %s:' % e.args[0])
        sys.exit(1)
    finally:
        if cursor is not None:
            cursor.close()
        if mem_cur is not None:
            mem_cur.close()

if os.path.isfile('static/hyper.db'):
    print('You are beyond hyper genesis')
else:
    # transaction processing
    hyper_cursor = None
    try:
        hyper_conn = sqlite3.connect('static/hyper.db')
        hyper_cursor = hyper_conn.cursor()
        hyper_cursor.execute('CREATE TABLE transactions (block_height INTEGER, timestamp, address, recipient, amount, signature, public_key, block_hash, fee, reward, operation, openfield)')
        hyper_cursor.execute(
            'INSERT INTO transactions VALUES (?,?,?,?,?,?,?,?,?,?,?,?)', ('1', timestamp, 'genesis', address, '0', signature_enc.decode('utf-8'), public_key_b64encoded, block_hash, 0, 1, 1, 'genesis')
        )  # Insert a row of data
        hyper_cursor.execute('CREATE TABLE misc (block_height INTEGER, difficulty TEXT)')
        hyper_cursor.execute('INSERT INTO misc (difficulty, block_height) VALUES ({},1)'.format(DIFFICULTY))
        # TODO: create indexes
        hyper_conn.commit()  # Save (commit) the changes

        hyper_conn.close()
        hyper_cursor = None

        print('Hyper Genesis created.')
    except sqlite3.Error as e:
        print('Error %s:' % e.args[0])
        sys.exit(1)
    finally:
        if hyper_cursor is not None:
            hyper_cursor.close()

if os.path.isfile('static/index.db'):
    print('Index already exists')
else:
    # transaction processing
    index_cursor = None
    try:
        index_conn = sqlite3.connect('static/index.db')
        index_cursor = index_conn.cursor()
        index_cursor.execute('CREATE TABLE tokens (block_height INTEGER, timestamp, token, address, recipient, txid, amount INTEGER)')
        index_cursor.execute('CREATE TABLE aliases (block_height INTEGER, address, alias)')
        index_cursor.execute('CREATE TABLE staking (block_height INTEGER, timestamp NUMERIC, address, balance, ip, port, pos_address)')
        # TODO: create indexes
        index_conn.commit()  # Save (commit) the changes

        index_conn.close()
        index_cursor = None

        print('Index table created.')
    except sqlite3.Error as e:
        print('Error %s:' % e.args[0])
        sys.exit(1)
    finally:
        if index_cursor is not None:
            index_cursor.close()



#!/usr/bin/python
# webtraffic.py
#
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (webtraffic.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
#
"""
.. module:: webtraffic.

A useful code to monitor program packets traffic in the Web browser
using local HTML server.
"""

from __future__ import absolute_import
import sys

try:
    from twisted.internet import reactor  # @UnresolvedImport
except:
    sys.exit('Error initializing twisted.internet.reactor in trafficstats.py')

from twisted.web import server, resource

#-------------------------------------------------------------------------------

#(total bytes, finished packets, failed packets, total packets)
_InboxPacketsCount = 0
_InboxByIDURL = {}
_InboxByHost = {}
_InboxByProto = {}
_InboxByType = {}

_OutboxPacketsCount = 0
_OutboxByIDURL = {}
_OutboxByHost = {}
_OutboxByProto = {}
_OutboxByType = {}

_WebListener = None

_DefaultReloadTimeout = 600

#-------------------------------------------------------------------------------


def init(root=None, path='traffic', port=9997):
    global _WebListener
    if root is not None:
        from bitdust.transport import callback
        callback.append_inbox_callback(inbox)
        callback.add_finish_file_sending_callback(outbox)
        root.putChild(path, TrafficPage())
        return
    if _WebListener:
        return
    root = resource.Resource()
    root.putChild('', TrafficPage())
    site = server.Site(root)
    try:
        _WebListener = reactor.listenTCP(port, site)
    except:
        from bitdust.logs import lg
        lg.exc()
        return
    from bitdust.transport import callback
    callback.append_inbox_callback(inbox)
    callback.add_finish_file_sending_callback(outbox)


def shutdown():
    global _WebListener
    if _WebListener:
        _WebListener.stopListening()
        del _WebListener
        _WebListener = None


def inbox_packets_count():
    global _InboxPacketsCount
    return _InboxPacketsCount


def inbox_by_idurl():
    global _InboxByIDURL
    return _InboxByIDURL


def inbox_by_host():
    global _InboxByHost
    return _InboxByHost


def inbox_by_proto():
    global _InboxByProto
    return _InboxByProto


def inbox_by_type():
    global _InboxByType
    return _InboxByType


def outbox_packets_count():
    global _OutboxPacketsCount
    return _OutboxPacketsCount


def outbox_by_idurl():
    global _OutboxByIDURL
    return _OutboxByIDURL


def outbox_by_host():
    global _OutboxByHost
    return _OutboxByHost


def outbox_by_proto():
    global _OutboxByProto
    return _OutboxByProto


def outbox_by_type():
    global _OutboxByType
    return _OutboxByType


def inbox(newpacket, info, status, error_message):
    global _InboxPacketsCount
    global _InboxByIDURL
    global _InboxByHost
    global _InboxByProto
    global _InboxByType

    if newpacket is None:
        return False

    byts = len(newpacket)
    idurl = newpacket.CreatorID
    host = '%s://%s' % (info.proto, info.host)
    typ = newpacket.Command

    if idurl not in _InboxByIDURL:
        _InboxByIDURL[idurl] = [0, 0, 0, 0]
    _InboxByIDURL[idurl][0] += byts
    if status == 'finished':
        _InboxByIDURL[idurl][1] += 1
    else:
        _InboxByIDURL[idurl][2] += 1
    _InboxByIDURL[idurl][3] += 1

    if host not in _InboxByHost:
        _InboxByHost[host] = [0, 0, 0, 0]
    _InboxByHost[host][0] += byts
    if status == 'finished':
        _InboxByHost[host][1] += 1
    else:
        _InboxByHost[host][2] += 1
    _InboxByHost[host][3] += 1

    if info.proto not in _InboxByProto:
        _InboxByProto[info.proto] = [0, 0, 0, 0]
    _InboxByProto[info.proto][0] += byts
    if status == 'finished':
        _InboxByProto[info.proto][1] += 1
    else:
        _InboxByProto[info.proto][2] += 1
    _InboxByProto[info.proto][3] += 1

    if typ not in _InboxByType:
        _InboxByType[typ] = [0, 0, 0, 0]
    _InboxByType[typ][0] += byts
    if status == 'finished':
        _InboxByType[typ][1] += 1
    else:
        _InboxByType[typ][2] += 1
    _InboxByType[typ][3] += 1

    _InboxPacketsCount += 1
    return False


def outbox(pkt_out, item, status, size, error_message):
    global _OutboxPacketsCount
    global _OutboxByIDURL
    global _OutboxByHost
    global _OutboxByProto
    global _OutboxByType

    byts = pkt_out.filesize
    idurl = pkt_out.remote_idurl
    host = '%s://%s' % (item.proto, item.host)
    typ = pkt_out.outpacket.Command

    if idurl not in _OutboxByIDURL:
        _OutboxByIDURL[idurl] = [0, 0, 0, 0]
    _OutboxByIDURL[idurl][0] += byts
    if status == 'finished':
        _OutboxByIDURL[idurl][1] += 1
    else:
        _OutboxByIDURL[idurl][2] += 1
    _OutboxByIDURL[idurl][3] += 1

    if host not in _OutboxByHost:
        _OutboxByHost[host] = [0, 0, 0, 0]
    _OutboxByHost[host][0] += byts
    if status == 'finished':
        _OutboxByHost[host][1] += 1
    else:
        _OutboxByHost[host][2] += 1
    _OutboxByHost[host][3] += 1

    if item.proto not in _OutboxByProto:
        _OutboxByProto[item.proto] = [0, 0, 0, 0]
    _OutboxByProto[item.proto][0] += byts
    if status == 'finished':
        _OutboxByProto[item.proto][1] += 1
    else:
        _OutboxByProto[item.proto][2] += 1
    _OutboxByProto[item.proto][3] += 1

    if typ not in _OutboxByType:
        _OutboxByType[typ] = [0, 0, 0, 0]
    _OutboxByType[typ][0] += byts
    if status == 'finished':
        _OutboxByType[typ][1] += 1
    else:
        _OutboxByType[typ][2] += 1
    _OutboxByType[typ][3] += 1

    _OutboxPacketsCount += 1
    return False


#-------------------------------------------------------------------------------


class TrafficPage(resource.Resource):
    header_html = '''<html><head>
<meta http-equiv="refresh" content="%(reload)s">
<title>Traffic</title></head>
<body bgcolor="#FFFFFF" text="#000000" link="#0000FF" vlink="#0000FF">
<form action="%(baseurl)s?" method="get">
<input size="4" name="reload" value="%(reload)s" />
<input type="submit" value="update" />
</form>
<a href="%(baseurl)s?reload=1&type=%(type)s&dir=%(dir)s">[1 sec.]</a>|
<a href="%(baseurl)s?reload=5&type=%(type)s&dir=%(dir)s">[5 sec.]</a>|
<a href="%(baseurl)s?reload=10&type=%(type)s&dir=%(dir)s">[10 sec.]</a>|
<a href="%(baseurl)s?reload=60&type=%(type)s&dir=%(dir)s">[60 sec.]</a>
<br>
<a href="%(baseurl)s?type=idurl&reload=%(reload)s&dir=%(dir)s">[by idurl]</a>|
<a href="%(baseurl)s?type=host&reload=%(reload)s&dir=%(dir)s">[by host]</a>|
<a href="%(baseurl)s?type=proto&reload=%(reload)s&dir=%(dir)s">[by proto]</a>
<a href="%(baseurl)s?type=type&reload=%(reload)s&dir=%(dir)s">[by type]</a>
<br>
<a href="%(baseurl)s?type=%(type)s&reload=%(reload)s&dir=in">[income traffic]</a>|
<a href="%(baseurl)s?type=%(type)s&reload=%(reload)s&dir=out">[outgoing traffic]</a>
<br>
<table><tr>
<td align=left>%(type)s
<td>total bytes
<td>total packets
<td>finished packets
<td>failed packets
'''

    def render(self, request):
        global _InboxPacketsCount
        global _InboxByIDURL
        global _InboxByHost
        global _InboxByProto
        global _OutboxPacketsCount
        global _OutboxByIDURL
        global _OutboxByHost
        global _OutboxByProto

        direction = request.args.get('dir', [''])[0]
        if direction not in ('in', 'out'):
            direction = 'in'
        typ = request.args.get('type', [''])[0]
        if typ not in ('idurl', 'host', 'proto', 'type'):
            typ = 'idurl'
        reloadS = request.args.get('reload', [''])[0]
        try:
            reloadV = int(reloadS)
        except:
            reloadV = _DefaultReloadTimeout

        d = {'type': typ, 'reload': str(reloadV), 'dir': direction, 'baseurl': request.path}
        out = self.header_html % d
        if direction == 'in':
            if typ == 'idurl':
                for i, v in _InboxByIDURL.items():
                    out += '<tr><td><a href="%s">%s</a><td>%d<td>%d<td>%d<td>%d\n' % (i, i, v[0], v[3], v[1], v[2])
            elif typ == 'host':
                for i, v in _InboxByHost.items():
                    out += '<tr><td>%s<td>%d<td>%d<td>%d<td>%d\n' % (i, v[0], v[3], v[1], v[2])
            elif typ == 'proto':
                for i, v in _InboxByProto.items():
                    out += '<tr><td>%s<td>%d<td>%d<td>%d<td>%d\n' % (i, v[0], v[3], v[1], v[2])
            elif typ == 'type':
                for i, v in _InboxByType.items():
                    out += '<tr><td>%s<td>%d<td>%d<td>%d<td>%d\n' % (i, v[0], v[3], v[1], v[2])
        else:
            if typ == 'idurl':
                for i, v in _OutboxByIDURL.items():
                    out += '<tr><td><a href="%s">%s</a><td>%d<td>%d<td>%d<td>%d\n' % (i, i, v[0], v[3], v[1], v[2])
            elif typ == 'host':
                for i, v in _OutboxByHost.items():
                    out += '<tr><td>%s<td>%d<td>%d<td>%d<td>%d\n' % (i, v[0], v[3], v[1], v[2])
            elif typ == 'proto':
                for i, v in _OutboxByProto.items():
                    out += '<tr><td>%s<td>%d<td>%d<td>%d<td>%d\n' % (i, v[0], v[3], v[1], v[2])
            elif typ == 'type':
                for i, v in _OutboxByType.items():
                    out += '<tr><td>%s<td>%d<td>%d<td>%d<td>%d\n' % (i, v[0], v[3], v[1], v[2])

        out += '</table>'
        if direction == 'in':
            out += '<p>total income packets: %d</p>' % _InboxPacketsCount
        else:
            out += '<p>total outgoing packets: %d</p>' % _OutboxPacketsCount
        out += '</body></html>'

        return out


#------------------------------------------------------------------------------

if __name__ == '__main__':
    init()
    reactor.run()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# Copyright 2011-2013 Codernity (http://codernity.com)
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


from __future__ import absolute_import
from threading import RLock

from bitdust_forks.CodernityDB3.env import cdb_environment

cdb_environment['mode'] = "threads"
cdb_environment['rlock_obj'] = RLock

from .database import Database

from functools import wraps
from types import FunctionType, MethodType

from bitdust_forks.CodernityDB3.database_safe_shared import th_safe_gen


class SuperLock(type):

    @staticmethod
    def wrapper(f):
        @wraps(f)
        def _inner(*args, **kwargs):
            db = args[0]
            with db.super_lock:
#                print('=>', f.__name__, repr(args[1:]))
                res = f(*args, **kwargs)
#                if db.opened:
#                    db.flush()
#                print('<=', f.__name__, repr(args[1:]))
                return res
        return _inner

    def __new__(cls, classname, bases, attr):
        new_attr = {}
        for base in bases:
            for b_attr in dir(base):
                a = getattr(base, b_attr, None)
                if isinstance(a, MethodType) and not b_attr.startswith('_'):
                    if b_attr == 'flush' or b_attr == 'flush_indexes':
                        pass
                    else:
                        # setattr(base, b_attr, SuperLock.wrapper(a))
                        new_attr[b_attr] = SuperLock.wrapper(a)
        for attr_name, attr_value in attr.items():
            if isinstance(attr_value, FunctionType) and not attr_name.startswith('_'):
                attr_value = SuperLock.wrapper(attr_value)
            new_attr[attr_name] = attr_value
        new_attr['super_lock'] = RLock()
        return type.__new__(cls, classname, bases, new_attr)


class SuperThreadSafeDatabase(Database, metaclass=SuperLock):  # pragma: no cover
    """
    Thread safe version that always allows single thread to use db.
    It adds the same lock for all methods, so only one operation can be
    performed in given time. Completely different implementation
    than ThreadSafe version (without super word)
    """

    def __init__(self, *args, **kwargs):
        super(SuperThreadSafeDatabase, self).__init__(*args, **kwargs)

    def __patch_index_gens(self, name):
        ind = self.indexes_names[name]
        for c in ('all', 'get_many'):
            m = getattr(ind, c)
            if getattr(ind, c + "_orig", None):
                return
            m_fixed = th_safe_gen.wrapper(m, name, c, self.super_lock)
            setattr(ind, c, m_fixed)
            setattr(ind, c + '_orig', m)

    def open(self, *args, **kwargs):
        res = super(SuperThreadSafeDatabase, self).open(*args, **kwargs)
        for name in self.indexes_names.keys():
            self.__patch_index_gens(name)
        return res

    def create(self, *args, **kwargs):
        res = super(SuperThreadSafeDatabase, self).create(*args, **kwargs)
        for name in self.indexes_names.keys():
            self.__patch_index_gens(name)
            return res

    def add_index(self, *args, **kwargs):
        res = super(SuperThreadSafeDatabase, self).add_index(*args, **kwargs)
        self.__patch_index_gens(res)
        return res

    def edit_index(self, *args, **kwargs):
        res = super(SuperThreadSafeDatabase, self).edit_index(*args, **kwargs)
        self.__patch_index_gens(res)
        return res



#!/usr/bin/python
# key.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (key.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
"""
.. module:: key.

Here is a bunch of cryptography methods used in all parts of the software.
BitDust uses PyCryptodome library: https://www.pycryptodome.org/
Our local key is always on hand.
Main thing here is to be able to use public keys in contacts to verify packets.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
from __future__ import print_function
from six.moves import range

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 4

#------------------------------------------------------------------------------

import os
import sys
import gc
import tempfile

#------------------------------------------------------------------------------

if __name__ == '__main__':
    import os.path as _p
    sys.path.insert(0, _p.abspath(_p.join(_p.dirname(_p.abspath(sys.argv[0])), '..')))

#------------------------------------------------------------------------------

from bitdust.logs import lg

from bitdust.system import bpio
from bitdust.system import local_fs

from bitdust.main import settings

from bitdust.crypt import rsa_key
from bitdust.crypt import hashes
from bitdust.crypt import cipher

#------------------------------------------------------------------------------

_MyKeyObject = None

#------------------------------------------------------------------------------


def InitMyKey(keyfilename=None):
    """
    This is core method. At first it check the Private Key in the memory, if it is already initialized it does nothing.
    The local key are placed in the `~/.bitdust/[network name]/metadata/mykeyfile`.

    If file "~/.bitdust/[network name]/metadata/mykeyfile_location" exists - it should contain the location
    of the mykeyfile. Useful to store Private Key on the USB flash.

    BitDust data dir is platform dependent:

        - Linux and Mac: ~/.bitdust
        - Windows XP: C:/Documents and Settings/[user]/.bitdust
        - Windows Vista, 7, 8: C:/Users/[user]/.bitdust
    Finally if target file exist - the Private Key will be loaded into memory.
    If file does not exist - the new key will be generated.
    The size for new key will be taken from settings.
    """
    global _MyKeyObject
    if _MyKeyObject is not None and _MyKeyObject.isReady():
        return False
    if not LoadMyKey(keyfilename):
        return False
        # GenerateNewKey(keyfilename)
    return True


def isMyKeyExists(keyfilename=None):
    if keyfilename is None:
        keyfilename = settings.KeyFileName()
    if os.path.exists(keyfilename + '_location'):
        newkeyfilename = bpio.ReadTextFile(keyfilename + '_location').strip()
        if os.path.exists(newkeyfilename):
            keyfilename = newkeyfilename
    return os.path.exists(keyfilename)


def LoadMyKey(keyfilename=None):
    global _MyKeyObject
    if keyfilename is None:
        keyfilename = settings.KeyFileName()
    if os.path.exists(keyfilename + '_location'):
        newkeyfilename = bpio.ReadTextFile(keyfilename + '_location').strip()
        if os.path.exists(newkeyfilename):
            keyfilename = newkeyfilename
    if not os.path.exists(keyfilename):
        return False
    _MyKeyObject = rsa_key.RSAKey()
    _MyKeyObject.fromFile(keyfilename)
    if _Debug:
        lg.out(_DebugLevel, 'key.InitMyKey loaded private key from %s' % (keyfilename))
    if not ValidateKey():
        if _Debug:
            lg.out(_DebugLevel, 'key.InitMyKey  private key is not valid: %s' % (keyfilename))
        return False
    return True


def GenerateNewKey(keyfilename=None):
    global _MyKeyObject
    if keyfilename is None:
        keyfilename = settings.KeyFileName()
    if os.path.exists(keyfilename + '_location'):
        newkeyfilename = bpio.ReadTextFile(keyfilename + '_location').strip()
        if os.path.exists(newkeyfilename):
            keyfilename = newkeyfilename
    if _Debug:
        lg.out(_DebugLevel, 'key.InitMyKey generate new private key')
    _MyKeyObject = rsa_key.RSAKey()
    _MyKeyObject.generate(settings.getPrivateKeySize())
    keystring = _MyKeyObject.toPrivateString()
    bpio.WriteTextFile(keyfilename, keystring)
    if _Debug:
        lg.out(_DebugLevel, '    wrote %d bytes to %s' % (len(keystring), keyfilename))
    del keystring
    gc.collect()


def ValidateKey():
    curkey = MyPrivateKeyObject()
    data256 = os.urandom(256)
    signature256 = curkey.sign(data256)
    return curkey.verify(signature256, data256)


def ForgetMyKey(keyfilename=None, erase_file=False, do_backup=False):
    """
    Remove Private Key from memory.
    """
    global _MyKeyObject
    if _MyKeyObject:
        _MyKeyObject.forget()
    _MyKeyObject = None
    if erase_file:
        if keyfilename is None:
            keyfilename = settings.KeyFileName()
        if do_backup:
            if os.path.isfile(keyfilename):
                current_pk_src = local_fs.ReadBinaryFile(keyfilename)
                if current_pk_src:
                    fd, fname = tempfile.mkstemp(prefix='mykeyfile_', dir=settings.MetaDataDir())
                    os.write(fd, current_pk_src)
                    os.close(fd)
                    lg.info('created backup copy of my private key in the file : %r' % fname)
        if os.path.isfile(keyfilename):
            os.remove(keyfilename)
            lg.info('local private key erased, deleted file : %r' % keyfilename)


def isMyKeyReady():
    """
    Check if the Key is already loaded into memory.
    """
    global _MyKeyObject
    return _MyKeyObject is not None and _MyKeyObject.isReady()


def MyPublicKey():
    """
    Return Public part of the Key as PEM string.
    """
    global _MyKeyObject
    InitMyKey()
    Result = _MyKeyObject.toPublicString()
    return Result


def MyPrivateKey():
    """
    Return Private part of the Key as PEM string.
    """
    global _MyKeyObject
    if not _MyKeyObject:
        InitMyKey()
    return _MyKeyObject.toPrivateString()


def MyPrivateKeyObject():
    """
    Return Private part of the Key as object.
    """
    global _MyKeyObject
    if not _MyKeyObject:
        InitMyKey()
    return _MyKeyObject


#------------------------------------------------------------------------------


def Sign(inp):
    """
    Sign some ``inp`` string with our Private Key, this calls PyCrypto method
    ``Crypto.PublicKey.RSA.sign``.
    """
    global _MyKeyObject
    if not _MyKeyObject:
        InitMyKey()
    result = _MyKeyObject.sign(inp)
    return result


def VerifySignature(pubkeystring, hashcode, signature):
    """
    Verify signature, this calls function ``Crypto.PublicKey.RSA.verify`` to
    verify.

    :param keystring: PublicKey in openssh format.
    :param hashcode: input data to verify, we use method ``Hash`` to prepare that.
    :param signature: string with signature to verify.

    Return True if signature is correct, otherwise False.
    """
    pub_key = rsa_key.RSAKey()
    pub_key.fromString(pubkeystring)
    result = pub_key.verify(signature, hashcode)
    return result


def Verify(ConIdentity, hashcode, signature):
    """
    This takes Public Key from user identity and calls ``VerifySignature``.

    :param ConIdentity: user's identity object'.
    """
    pubkey = ConIdentity.publickey
    Result = VerifySignature(pubkey, hashcode, signature)
    return Result


#------------------------------------------------------------------------------


def HashMD5(inp, hexdigest=False):
    """
    Use MD5 method to calculate the hash of ``inp`` string.

    However it seems it is not so safe anymore:
    http://natmchugh.blogspot.co.uk/2014/10/how-i-created-two-images-
    with-same-md5.html
    """
    return hashes.md5(inp, hexdigest=hexdigest)


def HashSHA(inp, hexdigest=False):
    """
    Use SHA1 method to calculate the hash of ``inp`` string.
    """
    return hashes.sha1(inp, hexdigest=hexdigest)


def HashSHA512(inp, hexdigest=False):
    return hashes.sha256(inp, hexdigest=hexdigest)


def Hash(inp, hexdigest=False):
    """
    Core function to calculate hash of ``inp`` string, right now it uses SHA1
    method.
    """
    return HashSHA(inp, hexdigest=hexdigest)


#------------------------------------------------------------------------------


def SessionKeyType():
    """
    Which crypto is used for session key.
    """
    return 'AES'


def NewSessionKey(session_key_type):
    """
    Return really random string for making AES cipher objects when needed.
    """
    return cipher.make_key(session_key_type)


#------------------------------------------------------------------------------


def EncryptWithSessionKey(session_key, inp, session_key_type):
    """
    Encrypt input string with Session Key.

    :param session_key: randomly generated session key
    :param inp: input string to encrypt
    """
    ret = cipher.encrypt_json(inp, session_key, session_key_type)
    return ret


def DecryptWithSessionKey(session_key, inp, session_key_type):
    """
    Decrypt string with given session key.

    :param session_key: a session key comes with the message in encrypted form,
        here it must be already decrypted
    :param inp: input string to decrypt
    """
    ret = cipher.decrypt_json(inp, session_key, session_key_type)
    return ret


#------------------------------------------------------------------------------


def EncryptOpenSSHPublicKey(pubkeystring, inp):
    """
    Encrypt ``inp`` string with given Public Key.
    """
    pub_key = rsa_key.RSAKey()
    pub_key.fromString(pubkeystring)
    result = pub_key.encrypt(inp)
    return result


def DecryptOpenSSHPrivateKey(privkeystring, inp):
    """
    Decrypt ``inp`` string with a Private Key provided as string in PEM format.
    """
    priv_key = rsa_key.RSAKey()
    priv_key.fromString(privkeystring)
    result = priv_key.decrypt(inp)
    return result


#------------------------------------------------------------------------------


def EncryptLocalPublicKey(inp):
    """
    This is just using local key, encrypt ``inp`` string.
    """
    global _MyKeyObject
    if not _MyKeyObject:
        InitMyKey()
    result = _MyKeyObject.encrypt(inp)
    return result


def DecryptLocalPrivateKey(inp):
    """
    Decrypt ``inp`` string with your Private Key.
    Here we decrypt with our local private key - so no argument for that.
    """
    global _MyKeyObject
    if not _MyKeyObject:
        InitMyKey()
    result = _MyKeyObject.decrypt(inp)
    return result


#------------------------------------------------------------------------------


def SpeedTest():
    """
    Some tests to check the performance.
    """
    import time
    dataSZ = 1024*640
    loops = 10
    packets = []
    dt = time.time()
    print('encrypt %d pieces of %d bytes' % (loops, dataSZ))
    for i in range(loops):
        Data = os.urandom(dataSZ)
        SessionKey = NewSessionKey(session_key_type=SessionKeyType())
        EncryptedSessionKey = EncryptLocalPublicKey(SessionKey)
        EncryptedData = EncryptWithSessionKey(SessionKey, Data, session_key_type=SessionKeyType())
        Signature = Sign(Hash(EncryptedData))
        packets.append((Data, len(Data), EncryptedSessionKey, EncryptedData, Signature))
        print('.', end=' ')
    print(time.time() - dt, 'seconds')

    dt = time.time()
    print('decrypt now')
    i = 0
    for Data, Length, EncryptedSessionKey, EncryptedData, Signature in packets:
        SessionKey = DecryptLocalPrivateKey(EncryptedSessionKey)
        paddedData = DecryptWithSessionKey(SessionKey, EncryptedData, session_key_type=SessionKeyType())
        newData = paddedData[:Length]
        if not VerifySignature(MyPublicKey(), Hash(EncryptedData), Signature):
            raise Exception()
        if newData != Data:
            raise Exception
        print('.', end=' ')
        # open(str(i), 'wb').write(EncryptedData)
        i += 1
    print(time.time() - dt, 'seconds')


#------------------------------------------------------------------------------

if __name__ == '__main__':
    bpio.init()
    lg.set_debug_level(18)
    settings.init()
    InitMyKey()
    SpeedTest()
    settings.shutdown()



#!/usr/bin/python
# config_defaults.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (config_defaults.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: config_defaults
"""


def reset(conf_obj):
    """
    Configure default values for all BitDust local settings inside ~/.bitdust/[network name]/config/ folder.

    Every option must have a default value, however there are exceptions possible.

    Here all items suppose to have some widget in UI to interact with human.
    Keep it simple and understandable.
    """
    from bitdust.lib import diskspace
    from bitdust.main import settings

    conf_obj.setDefaultValue('interface/api/auth-secret-enabled', 'true')

    conf_obj.setDefaultValue('interface/api/rest-http-enabled', 'true')
    conf_obj.setDefaultValue('interface/api/rest-http-port', settings.DefaultRESTHTTPPort())

    conf_obj.setDefaultValue('interface/api/web-socket-enabled', 'true')
    conf_obj.setDefaultValue('interface/api/web-socket-port', settings.DefaultWebSocketPort())

    conf_obj.setDefaultValue('interface/ftp/enabled', 'true')
    conf_obj.setDefaultValue('interface/ftp/port', settings.DefaultFTPPort())

    conf_obj.setDefaultValue('logs/api-enabled', 'false')
    conf_obj.setDefaultValue('logs/automat-transitions-enabled', 'false')
    conf_obj.setDefaultValue('logs/automat-events-enabled', 'false')
    conf_obj.setDefaultValue('logs/debug-level', settings.defaultDebugLevel())
    conf_obj.setDefaultValue('logs/memdebug-enabled', 'false')
    conf_obj.setDefaultValue('logs/memdebug-port', '9996')
    conf_obj.setDefaultValue('logs/memprofile-enabled', 'false')
    conf_obj.setDefaultValue('logs/packet-enabled', 'false')
    conf_obj.setDefaultValue('logs/stream-enabled', 'false')
    conf_obj.setDefaultValue('logs/stream-port', settings.DefaultWebLogPort())
    conf_obj.setDefaultValue('logs/traffic-enabled', 'false')
    conf_obj.setDefaultValue('logs/traffic-port', settings.DefaultWebTrafficPort())

    conf_obj.setDefaultValue('paths/backups', '')
    conf_obj.setDefaultValue('paths/customers', '')
    conf_obj.setDefaultValue('paths/messages', '')
    conf_obj.setDefaultValue('paths/receipts', '')
    conf_obj.setDefaultValue('paths/restore', '')

    conf_obj.setDefaultValue('personal/private-key-size', settings.DefaultPrivateKeySize())

    conf_obj.setDefaultValue('services/accountant/enabled', 'false')

    conf_obj.setDefaultValue('services/backup-db/enabled', 'true')

    conf_obj.setDefaultValue('services/backups/enabled', 'true')
    conf_obj.setDefaultValue('services/backups/block-size', diskspace.MakeStringFromBytes(settings.DefaultBackupBlockSize()))
    conf_obj.setDefaultValue('services/backups/max-block-size', diskspace.MakeStringFromBytes(settings.DefaultBackupMaxBlockSize()))
    conf_obj.setDefaultValue('services/backups/max-copies', '2')
    conf_obj.setDefaultValue('services/backups/keep-local-copies-enabled', 'true')
    conf_obj.setDefaultValue('services/backups/wait-suppliers-enabled', 'true')

    conf_obj.setDefaultValue('services/blockchain-id/enabled', 'true')

    conf_obj.setDefaultValue('services/blockchain-authority/enabled', 'false')
    conf_obj.setDefaultValue('services/blockchain-authority/registration-bonus-coins', 100)
    conf_obj.setDefaultValue('services/blockchain-authority/requests-reading-offset', 0)
    conf_obj.setDefaultValue('services/blockchain-authority/requests-reading-limit', 50)

    conf_obj.setDefaultValue('services/blockchain-explorer/enabled', 'false')
    conf_obj.setDefaultValue('services/blockchain-explorer/host', '127.0.0.1')
    conf_obj.setDefaultValue('services/blockchain-explorer/web-port', 19080)

    conf_obj.setDefaultValue('services/bismuth-blockchain/enabled', 'true')

    conf_obj.setDefaultValue('services/bismuth-node/enabled', 'false')
    conf_obj.setDefaultValue('services/bismuth-node/host', '127.0.0.1')
    conf_obj.setDefaultValue('services/bismuth-node/tcp-port', 15658)

    conf_obj.setDefaultValue('services/bismuth-wallet/enabled', 'true')

    conf_obj.setDefaultValue('services/bismuth-pool/enabled', 'false')
    conf_obj.setDefaultValue('services/bismuth-pool/host', '127.0.0.1')
    conf_obj.setDefaultValue('services/bismuth-pool/tcp-port', 18525)

    conf_obj.setDefaultValue('services/bismuth-miner/enabled', 'false')

    conf_obj.setDefaultValue('services/bismuth-identity/enabled', 'true')

    conf_obj.setDefaultValue('services/broadcasting/enabled', 'false')
    conf_obj.setDefaultValue('services/broadcasting/routing-enabled', 'false')
    conf_obj.setDefaultValue('services/broadcasting/max-broadcast-connections', '3')

    conf_obj.setDefaultValue('services/contract-chain/enabled', 'false')

    conf_obj.setDefaultValue('services/customer/enabled', 'true')
    conf_obj.setDefaultValue('services/customer/needed-space', diskspace.MakeStringFromBytes(settings.DefaultNeededBytes()))
    conf_obj.setDefaultValue('services/customer/suppliers-number', settings.DefaultDesiredSuppliers())

    conf_obj.setDefaultValue('services/customer-contracts/enabled', 'true')

    conf_obj.setDefaultValue('services/customer-family/enabled', 'true')

    conf_obj.setDefaultValue('services/customer-patrol/enabled', 'true')
    conf_obj.setDefaultValue('services/customer-patrol/customer-idle-days', 14)

    conf_obj.setDefaultValue('services/customer-support/enabled', 'true')

    conf_obj.setDefaultValue('services/data-disintegration/enabled', 'true')

    conf_obj.setDefaultValue('services/data-motion/enabled', 'true')
    conf_obj.setDefaultValue('services/data-motion/supplier-request-queue-size', 4)
    conf_obj.setDefaultValue('services/data-motion/supplier-sending-queue-size', 4)

    conf_obj.setDefaultValue('services/entangled-dht/enabled', 'true')
    conf_obj.setDefaultValue('services/entangled-dht/udp-port', settings.DefaultDHTPort())
    conf_obj.setDefaultValue('services/entangled-dht/known-nodes', '')
    conf_obj.setDefaultValue('services/entangled-dht/attached-layers', '')

    conf_obj.setDefaultValue('services/employer/enabled', 'true')
    conf_obj.setDefaultValue('services/employer/replace-critically-offline-enabled', 'true')
    conf_obj.setDefaultValue('services/employer/candidates', '')

    conf_obj.setDefaultValue('services/gateway/enabled', 'true')
    conf_obj.setDefaultValue('services/gateway/p2p-timeout', 15)

    conf_obj.setDefaultValue('services/http-connections/enabled', 'false')
    conf_obj.setDefaultValue('services/http-connections/http-port', settings.DefaultHTTPPort())

    conf_obj.setDefaultValue('services/http-transport/enabled', 'false')  # not done yet
    conf_obj.setDefaultValue('services/http-transport/receiving-enabled', 'true')
    conf_obj.setDefaultValue('services/http-transport/sending-enabled', 'true')
    conf_obj.setDefaultValue('services/http-transport/priority', 50)

    conf_obj.setDefaultValue('services/identity-server/enabled', 'false')
    conf_obj.setDefaultValue('services/identity-server/host', '')
    conf_obj.setDefaultValue('services/identity-server/tcp-port', settings.IdentityServerPort())
    conf_obj.setDefaultValue('services/identity-server/web-port', settings.IdentityWebPort())

    conf_obj.setDefaultValue('services/identity-propagate/enabled', 'true')
    conf_obj.setDefaultValue('services/identity-propagate/known-servers', '')
    conf_obj.setDefaultValue('services/identity-propagate/preferred-servers', '')
    conf_obj.setDefaultValue('services/identity-propagate/min-servers', 2)
    conf_obj.setDefaultValue('services/identity-propagate/max-servers', 5)
    conf_obj.setDefaultValue('services/identity-propagate/automatic-rotate-enabled', 'true')
    conf_obj.setDefaultValue('services/identity-propagate/health-check-interval-seconds', 60*5)

    conf_obj.setDefaultValue('services/joint-postman/enabled', 'true')

    conf_obj.setDefaultValue('services/ip-port-responder/enabled', 'true')

    conf_obj.setDefaultValue('services/keys-registry/enabled', 'true')

    conf_obj.setDefaultValue('services/keys-storage/enabled', 'true')
    conf_obj.setDefaultValue('services/keys-storage/reset-unreliable-backup-copies', 'true')

    conf_obj.setDefaultValue('services/list-files/enabled', 'true')

    conf_obj.setDefaultValue('services/message-broker/enabled', 'false')
    conf_obj.setDefaultValue('services/message-broker/archive-chunk-size', 100)
    conf_obj.setDefaultValue('services/message-broker/message-ack-timeout', 30)
    conf_obj.setDefaultValue('services/message-broker/broker-negotiate-ack-timeout', 30)
    conf_obj.setDefaultValue('services/message-broker/preferred-brokers', '')

    conf_obj.setDefaultValue('services/message-history/enabled', 'true')

    conf_obj.setDefaultValue('services/miner/enabled', 'false')

    conf_obj.setDefaultValue('services/my-data/enabled', 'true')

    conf_obj.setDefaultValue('services/my-ip-port/enabled', 'true')

    conf_obj.setDefaultValue('services/network/enabled', 'true')
    conf_obj.setDefaultValue('services/network/proxy/enabled', 'false')
    conf_obj.setDefaultValue('services/network/proxy/host', '')
    conf_obj.setDefaultValue('services/network/proxy/password', '')
    conf_obj.setDefaultValue('services/network/proxy/port', '')
    conf_obj.setDefaultValue('services/network/proxy/ssl', 'false')
    conf_obj.setDefaultValue('services/network/proxy/username', '')
    conf_obj.setDefaultValue('services/network/receive-limit', settings.DefaultBandwidthInLimit())
    conf_obj.setDefaultValue('services/network/send-limit', settings.DefaultBandwidthOutLimit())

    conf_obj.setDefaultValue('services/nodes-lookup/enabled', 'true')

    conf_obj.setDefaultValue('services/p2p-hookups/enabled', 'true')

    conf_obj.setDefaultValue('services/p2p-notifications/enabled', 'true')

    conf_obj.setDefaultValue('services/personal-messages/enabled', 'true')

    conf_obj.setDefaultValue('services/private-groups/enabled', 'true')
    conf_obj.setDefaultValue('services/private-groups/message-ack-timeout', 30)
    conf_obj.setDefaultValue('services/private-groups/broker-connect-timeout', 120)
    conf_obj.setDefaultValue('services/private-groups/preferred-brokers', '')

    conf_obj.setDefaultValue('services/private-messages/enabled', 'true')
    conf_obj.setDefaultValue('services/private-messages/acknowledge-unread-messages-enabled', 'true')

    conf_obj.setDefaultValue('services/proxy-server/enabled', 'false')
    conf_obj.setDefaultValue('services/proxy-server/routes-limit', 20)
    conf_obj.setDefaultValue('services/proxy-server/current-routes', '{}')

    conf_obj.setDefaultValue('services/proxy-transport/enabled', 'true')
    conf_obj.setDefaultValue('services/proxy-transport/sending-enabled', 'true')
    conf_obj.setDefaultValue('services/proxy-transport/receiving-enabled', 'true')
    conf_obj.setDefaultValue('services/proxy-transport/priority', 100)
    conf_obj.setDefaultValue('services/proxy-transport/preferred-routers', '')
    # conf_obj.setDefaultValue('services/proxy-transport/router-lifetime-seconds', 600)
    # TODO: those two settings needs to be removed.
    # if service require storing locally a value which user should not modify we better move it to another place
    # in the future we can split those local data files in more structural way and move into
    # ~/.bitdust/[network name]/services/*/ sub folders
    conf_obj.setDefaultValue('services/proxy-transport/my-original-identity', '')
    conf_obj.setDefaultValue('services/proxy-transport/current-router', '')

    conf_obj.setDefaultValue('services/rebuilding/enabled', 'true')

    conf_obj.setDefaultValue('services/restores/enabled', 'true')

    conf_obj.setDefaultValue('services/shared-data/enabled', 'true')

    conf_obj.setDefaultValue('services/supplier/enabled', 'true')
    conf_obj.setDefaultValue('services/supplier/donated-space', diskspace.MakeStringFromBytes(settings.DefaultDonatedBytes()))

    conf_obj.setDefaultValue('services/supplier-contracts/enabled', 'true')
    conf_obj.setDefaultValue('services/supplier-contracts/initial-duration-hours', 6)
    conf_obj.setDefaultValue('services/supplier-contracts/duration-raise-factor', 2.0)

    conf_obj.setDefaultValue('services/tcp-connections/enabled', 'true')
    conf_obj.setDefaultValue('services/tcp-connections/tcp-port', settings.DefaultTCPPort())
    conf_obj.setDefaultValue('services/tcp-connections/upnp-enabled', 'true')

    conf_obj.setDefaultValue('services/tcp-transport/enabled', 'true')
    conf_obj.setDefaultValue('services/tcp-transport/receiving-enabled', 'true')
    conf_obj.setDefaultValue('services/tcp-transport/sending-enabled', 'true')
    conf_obj.setDefaultValue('services/tcp-transport/priority', 10)

    conf_obj.setDefaultValue('services/udp-datagrams/enabled', 'true')
    conf_obj.setDefaultValue('services/udp-datagrams/udp-port', settings.DefaultUDPPort())

    # TODO: UDP transport was temporary switched off
    conf_obj.setDefaultValue('services/udp-transport/enabled', 'false')
    conf_obj.setDefaultValue('services/udp-transport/receiving-enabled', 'true')
    conf_obj.setDefaultValue('services/udp-transport/sending-enabled', 'true')
    conf_obj.setDefaultValue('services/udp-transport/priority', 20)



#!/usr/bin/python
# service_my_ip_port.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (service_my_ip_port.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
..

module:: service_my_ip_port
"""

from __future__ import absolute_import
from bitdust.services.local_service import LocalService


def create_service():
    return MyIPPortService()


class MyIPPortService(LocalService):

    service_name = 'service_my_ip_port'
    config_path = 'services/my-ip-port/enabled'
    start_suspended = True

    def init(self):
        self._my_address = None

    def dependent_on(self):
        return [
            'service_entangled_dht',
            'service_udp_datagrams',
        ]

    def start(self):
        from bitdust.stun import stun_client
        from bitdust.main import settings
        from bitdust.lib import misc
        stun_client.A('init', settings.getUDPPort())
        known_external_ip = misc.readExternalIP()
        if not known_external_ip or known_external_ip == '127.0.0.1':
            self._do_stun()
        return True

    def stop(self):
        from bitdust.stun import stun_client
        stun_client.A('shutdown')
        return True

    def on_suspend(self, *args, **kwargs):
        return True

    def on_resume(self, *args, **kwargs):
        from bitdust.stun import stun_client
        if not stun_client.A() or stun_client.A().state in [
            'STOPPED',
        ]:
            stun_client.A().dropMyExternalAddress()
            stun_client.A('start')
        return True

    def _do_stun(self):
        from bitdust.stun import stun_client
        stun_client.A().dropMyExternalAddress()
        stun_client.A('start', self._on_stun_result)

    def _on_stun_result(self, stun_result, nat_type, my_ip, details):
        from bitdust.logs import lg
        from twisted.internet import reactor  # @UnresolvedImport
        if stun_result != 'stun-success' or not my_ip or my_ip == '127.0.0.1':
            lg.warn('stun my external IP failed, retry after 10 seconds')
            reactor.callLater(10, self._do_stun)  # @UndefinedVariable
        else:
            lg.info('stun success  nat_type=%r, my_ip=%r, details=%r' % (nat_type, my_ip, details))



#!/usr/bin/python
# automat.py
#
# Copyright (C) 2008 Veselin Penev, https://bitdust.io
#
# This file (automat.py) is part of BitDust Software.
#
# BitDust is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# BitDust Software is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU Affero General Public License for more details.
#
# You should have received a copy of the GNU Affero General Public License
# along with BitDust Software.  If not, see <http://www.gnu.org/licenses/>.
#
# Please contact us if you have any questions at bitdust.io@gmail.com
#
#
#
#
"""
.. module:: automat.

This is the base class for State Machine.
The BitDust project is developing in principles of
`Automata-based programming <http://en.wikipedia.org/wiki/Automata-based_programming>`_.

This is a programming paradigm in which the program or its part is thought of as a model of a
`finite state machine <http://en.wikipedia.org/wiki/Finite_state_machine>`_ or any other formal automaton.

Its defining characteristic is the use of finite state machines to
`describe program behavior <http://en.wikipedia.org/wiki/State_diagram>`_.

The transition graphs of state machines are used in all stages of software development:
- specification,
- implementation,
- debugging and
- documentation.

You can see Transition graph for all BitDust state machines in the file
`automats.pdf <https://bitdust.io/automats.pdf>`_,
MS Visio, 'editable' version:
`automats.vsd <https://bitdust.io/automats.vsd>`_,
stencils is here: `automats.vss <https://bitdust.io/automats.vss>`_

A small tool called `visio2python <http://code.google.com/p/visio2python/>`_
was written to simplify working with the BitDust project.

It can translate transition graphs created in Microsoft Visio into Python code.

Automata-Based Programming technology was introduced by Anatoly Shalyto in 1991 and Switch-technology was
developed to support automata-based programming.
Automata-Based Programming is considered to be rather general purpose program development methodology
than just another one finite state machine implementation.
Anatoly Shalyto is the former of
`Foundation for Open Project Documentation <http://en.wikipedia.org/wiki/Foundation_for_Open_Project_Documentation>`_.

Read more about Switch-technology on the Saint-Petersburg National Research University
of Information Technologies, Mechanics and Optics, Programming Technologies Department
`Page <http://is.ifmo.ru/english>`_.
"""

#------------------------------------------------------------------------------

from __future__ import absolute_import
import sys
import time
import traceback
from io import open

from twisted.internet import reactor  # @UnresolvedImport
from twisted.internet.task import LoopingCall  #@UnresolvedImport
from twisted.internet.defer import Deferred, fail  #@UnresolvedImport

#------------------------------------------------------------------------------

_Debug = False
_DebugLevel = 10

#------------------------------------------------------------------------------

_LogFile = None  # : This is to have a separated Log file for state machines logs
_LogFilename = None
_LogsCount = 0  # : If not zero - it will print time since that value, not system time
_LogOutputHandler = None
_LogExceptionsHandler = None
_LifeBeginsTime = 0
_GlobalLogEvents = False
_GlobalLogTransitions = False

#------------------------------------------------------------------------------

_Counter = 0  # : Increment by one for every new object, the idea is to keep unique ID's in the index
_Index = {}  # : Index dictionary, unique id (string) to index (int)
_Objects = {}  # : Objects dictionary to store all state machines objects
_StateChangedCallback = None  # : Called when some state were changed

#------------------------------------------------------------------------------


def init():
    pass


def shutdown():
    global _Counter
    global _Index
    global _Objects
    global _StateChangedCallback
    LifeBegins(0)
    CloseLogFile()
    SetGlobalLogEvents()
    SetGlobalLogTransitions()
    SetExceptionsHandler(None)
    SetLogOutputHandler(None)
    _StateChangedCallback = None
    _Index.clear()
    _Objects.clear()
    _Counter = 0


#------------------------------------------------------------------------------


def get_new_index():
    """
    Just get the current index and increase by one.
    """
    global _Counter
    _Counter += 1
    return _Counter


def create_index(name):
    """
    Generate unique ID, and put it into Index dict, increment counter.
    """
    global _Index
    automatid = name
    if id in _Index:
        i = 1
        while automatid + '(' + str(i) + ')' in _Index:
            i += 1
        automatid = name + '(' + str(i) + ')'
    _Index[automatid] = get_new_index()
    return automatid, _Index[automatid]


def erase_index(automatid):
    """
    Removes given unique automat ID from Index dict and returns its index number or None if ID was not found.
    """
    global _Index
    if _Index is None:
        return None
    return _Index.pop(automatid, None)


def set_object(index, obj):
    """
    Put object for that index into memory.
    """
    global _Objects
    _Objects[index] = obj


def clear_object(index):
    """
    Clear object with given index from memory.
    """
    global _Objects
    if _Objects is None:
        return
    return _Objects.pop(index, None)


def objects():
    """
    Get all state machines stored in memory.
    """
    global _Objects
    return _Objects


def index():
    """
    Get all indexed state machines - should match with objects()
    """
    global _Index
    return _Index


def find(name):
    """
    Find state machine by name, this method will iterate all registered state machines
    and match "name" field to find the result.
    Return a list of indexes, then you can access those automats via `objects()` dict.
    """
    results = []
    for sm in objects().values():
        if sm.name == name:
            results.append(sm.index)
    return results


def by_index(index):
    """
    Returns state machine instance with given ``index`` if exists, otherwise returns `None`.
    """
    return objects().get(index, None)


def by_id(automat_id):
    """
    Returns state machine instance with given ``id`` if exists, otherwise returns `None`.
    """
    _index = index().get(automat_id, None)
    if _index is None:
        return None
    return by_index(_index)


def communicate(index, event, *args, **kwargs):
    """
    You can pass an event to any state machine - select by its ``index``.
    Use ``arg`` to pass extra data the conditions and actions methods.
    This method creates a Deferred object, pass it as a parameter with ``event``
    into the state machine and return that defer to outside - to catch result.
    In the action method you must call ``callback`` or ``errback`` to pass result.
    """
    A = by_index(index)
    if not A:
        return fail(Exception('state machine with index %d not exist' % index))
    d = Deferred()
    args = tuple(list(args) + [d])
    A.automat(event, *args, **kwargs)
    return d


#------------------------------------------------------------------------------


def SetStateChangedCallback(cb):
    """
    Set callback to be fired when any state machine changes its state Callback
    parameters are::

    cb(index, id, name, new state)
    """
    global _StateChangedCallback
    _StateChangedCallback = cb


def SetLogOutputHandler(cb):
    """
    Set callback to be fired when a log line is about to be printed from any state machine
    parameters are::

    cb(debug_level, message)
    """
    global _LogOutputHandler
    _LogOutputHandler = cb


def SetExceptionsHandler(cb):
    """
    Set callback to be fired when exception is caught from any state machine
    parameters are::

    cb(msg, exc_info)
    """
    global _LogExceptionsHandler
    _LogExceptionsHandler = cb


def RedirectLogFile(stream):
    """
    You can simple send all output to the stdout: import sys; RedirectLogFile(sys.stdout)
    """
    global _LogFile
    _LogFile = stream


def OpenLogFile(filename):
    """
    Open a file to write logs from all state machines.

    Very useful during debug.
    """
    global _LogFile
    global _LogFilename
    if _LogFile:
        return
    _LogFilename = filename
    try:
        _LogFile = open(_LogFilename, 'w')
    except:
        _LogFile = None


def CloseLogFile():
    """
    Close the current log file, you can than open it again.
    """
    global _LogFile
    if not _LogFile:
        return
    _LogFile.flush()
    _LogFile.close()
    _LogFile = None
    _LogFilename = None


def LifeBegins(when=None):
    """
    Call that function during program start up to print relative time in the
    logs, not absolute.
    """
    global _LifeBeginsTime
    if when is not None:
        _LifeBeginsTime = when
    else:
        _LifeBeginsTime = time.time()


def SetGlobalLogEvents(value=False):
    global _GlobalLogEvents
    _GlobalLogEvents = value


def SetGlobalLogTransitions(value=False):
    global _GlobalLogTransitions
    _GlobalLogTransitions = value


#------------------------------------------------------------------------------


class Automat(object):
    """
    Base class of the State Machine Object.

    You need to subclass this class and override the method ``A(event, *args, **kwargs)``.
    Constructor needs the ``name`` of the state machine and the
    beginning ``state``. At first it generate an unique ``id`` and new
    ``index`` value. You can use ``init()`` method in the subclass to
    call some code at start. Finally put the new object into the memory
    with given index - it is placed into ``objects()`` dictionary. To
    remove the instance call ``destroy()`` method.
    """

    state = 'NOT_EXIST'
    """
    This is a string representing current Machine state, must be set in the constructor.
    ``NOT_EXIST`` indicates that this machine is not created yet.
    A blank state is a fundamental mistake!
    """

    timers = {}
    """
    A dictionary of timer events fired at specified intervals when machine rich given state:
          timers = {'timer-60sec':     (60,     ['STATE_A',]),
                    'timer-3min':      (60*3,   ['STATE_B', 'STATE_C',]), }
    """

    instant_timers = False
    """
    Set this to True and timers will not skip first iteration.
    See method self.startTimers().
    """

    fast = True
    """
    By default, incoming state machine events are passing via Twisted reactor like that:

        reactor.callLater(0, self.event, 'event-01', (arg1, arg2, ... ))

    But when ``fast = True`` event will be passed directly as function call of the state machine:

        self.event('event-01', (arg1, arg2, ... ))
    """

    post = False
    """
    Sometimes need to set the new state AFTER finish all actions.
    Set ``post = True`` to call ``self.state = <newstate>``
    in the ``self.event()`` method, but not in the ``self.A()`` method.
    You also must set that flag in the MS Visio document and rebuild the code:
    put ``[post]`` string into the last line of the LABEL shape.
    """
    def __init__(self, name, state, debug_level=_DebugLevel, log_events=_Debug, log_transitions=_Debug, publish_events=False, publish_event_state_not_changed=False, publish_fast=True, **kwargs):
        self.id, self.index = create_index(name)
        self.name = name
        self.state = state
        self.debug_level = debug_level
        self.log_events = log_events
        self.log_transitions = log_transitions
        self._timers = {}
        self._state_callbacks = {}
        self._callbacks_before_die = {}
        try:
            self.init(**kwargs)
        except Exception as exc:
            self.exc(msg='Exception in {}:{} automat init(), state is {}: {}'.format(self.id, self.name, self.state, exc))
            raise exc
        self.startTimers()
        self.register()
        self.publish_events = publish_events
        self.publish_event_state_not_changed = publish_event_state_not_changed
        self.publish_fast = publish_fast
        if _GlobalLogTransitions or self.log_transitions:
            self.log(self.debug_level, 'CREATED AUTOMAT with index %d, total running %d' % (self.index, len(objects())))

    def __del__(self):
        """
        Calls state changed callback and removes state machine from the index.
        """
        global _StateChangedCallback
        global _GlobalLogTransitions
        global _LogFile
        if self is None:
            self.log(self.debug_level, 'Some crazy stuff happens?')
            return
        automatid = self.id
        name = self.name
        index = self.index
        if _StateChangedCallback is not None:
            _StateChangedCallback(index, automatid, name, '')
        debug_level = self.debug_level or 0
        if erase_index:
            erase_index(automatid)
        if _GlobalLogTransitions or self.log_transitions:
            if _LogFile:
                self.log(debug_level, 'DESTROYED AUTOMAT with index %d, total running %d' % (index, len(objects())))

    def __repr__(self):
        """
        Will return something like: "network_connector(CONNECTED)".
        """
        return '%s(%s)' % (self.id, self.state)

    def A(self, event, *args, **kwargs):
        """
        Must define this method in subclass.

        This is the core method of the SWITCH-technology. I am using
        ``visio2python`` (created by me) to generate Python code from MS
        Visio drawing.
        """
        raise NotImplementedError

    def init(self, **kwargs):
        """
        Define this method in subclass to execute some code when creating a
        new instance of Automat class.
        """

    def shutdown(self):
        """
        Define this method in subclass to execute some code when closing an
        existing instance of Automat class.
        """

    def register(self):
        """
        Put reference to this automat instance into a global dictionary.
        """
        set_object(self.index, self)
        return self.index

    def unregister(self):
        """
        Removes reference to this instance from global dictionary tracking all state machines.
        """
        clear_object(self.index)
        return True

    def destroy(self, dead_state=None):
        """
        Call this method to remove the state machine from the ``objects()``
        dictionary and delete that instance.

        Be sure to not have any existing references on that instance so
        destructor will be called immediately.
        """
        self.shutdown()
        if dead_state:
            self.state = dead_state
        self._callbacks_before_die = self._state_callbacks.copy()
        self._state_callbacks.clear()
        self.stopTimers()
        return self.unregister()

    def state_changed(self, oldstate, newstate, event, *args, **kwargs):
        """
        Redefine this method in subclass to be able to catch the moment
        immediately after automat's state were changed.
        """

    def state_not_changed(self, curstate, event, *args, **kwargs):
        """
        Redefine this method in subclass if you want to do some actions
        immediately after processing the event, which did not change the
        automat's state.
        """

    def communicate(self, event_string, *args, **kwargs):
        """
        Use ``arg`` to pass extra data the conditions and actions methods. This
        method creates a Deferred object, pass it as a parameter with ``event``

        into the state machine and return that defer to outside - to catch result.
        In the action method you must call ``callback`` or ``errback`` to pass result.

        See ``addStateChangedCallback()`` for more advanced interactions/callbacks.
        """
        d = Deferred()
        if not args:
            args = tuple()
        args = tuple(list(args) + [d])
        self.automat(event_string, args)
        return d

    def automat(self, event, *args, **kwargs):
        """
        Call it like this::

            machineA.automat('init', arguments)

        to send some ``event`` to the State Machine Object.
        You can attach parameters to that event with ``arguments`` tuple.
        If ``self.fast=False`` - the ``self.A()`` method will be executed in delayed call.
        """
        _fast = kwargs.pop('fast', False)
        if self.fast or _fast:
            self.event(event, *args, **kwargs)
        else:
            delay = kwargs.pop('delay', 0)
            reactor.callLater(delay, self.event, event, *args, **kwargs)  # @UndefinedVariable
        return self

    def event(self, event, *args, **kwargs):
        """
        You can call ``event()`` directly to execute ``self.A()`` immediately,
        but preferred way is too call ``automat()`` method.

        Use ``fast = True`` flag to skip call to reactor.callLater(0, self.event, ...).
        """
        global _StateChangedCallback
        if _GlobalLogEvents or self.log_events:
            if self.log_events or not event.startswith('timer-'):
                self.log(self.debug_level, '%s fired with event "%s"' % (
                    repr(self),
                    event,
                ))
        old_state = self.state
        if self.post:
            try:
                new_state = self.A(event, *args, **kwargs)
            except Exception as exc:
                self.exc(msg='Exception in {}:{} automat, state is {}, event="{}" : {}'.format(self.id, self.name, self.state, event, exc))
                return self
            self.state = new_state
        else:
            try:
                self.A(event, *args, **kwargs)
            except Exception as exc:
                self.exc(msg='Exception in {}:{} automat, state is {}, event="{}" : {}'.format(self.id, self.name, self.state, event, exc))
                return self
            new_state = self.state
        if old_state != new_state:
            if _GlobalLogTransitions or self.log_transitions:
                self.log(self.debug_level, '%s(%s): (%s)->(%s)' % (
                    repr(self),
                    event,
                    old_state,
                    new_state,
                ))
            self.stopTimers()
            self.state_changed(old_state, new_state, event, *args, **kwargs)
            if self.publish_events:
                self.pushEvent(old_state, new_state, event)
            self.startTimers()
            if _StateChangedCallback is not None:
                _StateChangedCallback(self.index, self.id, self.name, new_state)
        else:
            self.state_not_changed(self.state, event, *args, **kwargs)
            if self.publish_events:
                if self.publish_event_state_not_changed:
                    self.pushEvent(old_state, new_state, event)
        self.executeStateChangedCallbacks(old_state, new_state, event, *args, **kwargs)
        return self

    def timerEvent(self, name, interval):
        """
        This method fires the timer events.
        """
        try:
            if name in self.timers and self.state in self.timers[name][1]:
                self.automat(name)
            else:
                if _GlobalLogEvents or self.log_events:
                    self.log(self.debug_level, '%s.timerEvent ERROR timer %s not found in self.timers' % (str(self), name))
        except:
            self.exc()

    def stopTimers(self):
        """
        Stop all state machine timers.
        """
        for name, timer in self._timers.items():  # @UnusedVariable
            if timer.running:
                timer.stop()
        self._timers.clear()

    def startTimers(self):
        """
        Start all state machine timers.
        """
        for name, (interval, states) in self.timers.items():
            if len(states) > 0 and self.state not in states:
                continue
            self._timers[name] = LoopingCall(self.timerEvent, name, interval)
            self._timers[name].start(interval, self.instant_timers)

    def restartTimers(self):
        """
        Restart all state machine timers.

        When state is changed - all internal timers is restarted.
        You can use external timers if you do not need that, call::

            machineA.automat('timer-1sec')

        to fire timer event from outside.
        """
        self.stopTimers()
        self.startTimers()

    def getTimers(self):
        """
        Get internal timers dictionary.
        """
        return self._timers

    def to_json(self, short=True):
        if short:
            return {
                'index': self.index,
                'id': self.id,
                'name': self.__class__.__name__,
                'state': self.state,
            }
        return {
            'index': self.index,
            'id': self.id,
            'name': self.__class__.__name__,
            'state': self.state,
            'repr': repr(self),
            'timers': (','.join(list(self.getTimers().keys()))),
            'events': self.publish_events,
        }

    def exc(self, msg='', to_logfile=False, exc_type=None, exc_value=None, exc_traceback=None):
        """
        Print exception in stdout, optionally to log file.
        """
        global _LogExceptionsHandler
        global _LogFile
        _t = exc_type
        _v = exc_value
        _tb = exc_traceback
        if exc_type is None or exc_value is None or exc_traceback is None:
            _t, _v, _tb = sys.exc_info()
        exc_type = _t if exc_type is None else exc_type
        exc_value = _v if exc_value is None else exc_value
        exc_traceback = _tb if exc_traceback is None else exc_traceback
        e = ''
        if exc_value is not None or exc_traceback is not None:
            e = traceback.format_exception(exc_type, value=exc_value, tb=exc_traceback)
        else:
            e = traceback.format_exc()
        if to_logfile and _LogFile is not None:
            if msg:
                self.log(0, msg)
            self.log(0, e)
        if _LogExceptionsHandler is not None:
            _LogExceptionsHandler(msg=msg, exc_info=(
                exc_type,
                exc_value,
                exc_traceback,
            ))

    def log(self, level, text):
        """
        Print log message.

        See ``OpenLogFile()`` and ``CloseLogFile()`` methods.
        """
        global _LogFile
        global _LogFilename
        global _LogsCount
        global _LifeBeginsTime
        global _LogOutputHandler
        if not text.startswith(self.name):
            text = '%s(): %s' % (
                self.name,
                text,
            )
        if _LogOutputHandler is not None:
            _LogOutputHandler(level, text)
        else:
            if _LogFile is not None:
                if _LogsCount > 100000 and _LogFilename:
                    # very simple log rotation
                    _LogFile.close()
                    _LogFile = open(_LogFilename, 'w')
                    _LogsCount = 0
                s = ' '*level + text + '\n'
                tm_str = time.strftime('%H:%M:%S')
                if _LifeBeginsTime != 0:
                    dt = time.time() - _LifeBeginsTime
                    mn = dt // 60
                    sc = dt - mn*60
                    tm_str += ('/%02d:%02d.%02d' % (mn, sc, (sc - int(sc))*100))
                s = tm_str + s
                if sys.version_info[0] == 3:
                    if not isinstance(s, str):
                        s = s.decode('utf-8')
                else:
                    if not isinstance(s, unicode):  # @UndefinedVariable
                        s = s.decode('utf-8')
                try:
                    _LogFile.write(s)
                    _LogFile.flush()
                except:
                    pass
                _LogsCount += 1

    def addStateChangedCallback(self, cb, oldstate=None, newstate=None, callback_id=None):
        """
        You can add a callback function to be executed when state machine
        reaches particular condition, it will be called with such arguments:

            cb(oldstate, newstate, event_string, *args, **kwargs)

        For example, method_B() will be called when machine_A become "ONLINE":

            machine_A.addStateChangedCallback(method_B, None, "ONLINE")

        If you set "None" to both arguments,
        the callback will be executed every time when the state gets changed:

            machineB.addStateChangedCallback(method_B)

        """
        key = (
            oldstate,
            newstate,
        )
        if key not in self._state_callbacks:
            self._state_callbacks[key] = []
        if cb not in self._state_callbacks[key]:
            self._state_callbacks[key].append((callback_id, cb))
        return True

    def removeStateChangedCallback(self, cb=None, callback_id=None):
        """
        Remove given callback from the state machine.
        """
        removed_count = 0
        for key in list(self._state_callbacks.keys()):
            cb_list = self._state_callbacks[key]
            for cb_tupl in cb_list:
                cb_id_, cb_ = cb_tupl
                if cb and cb == cb_:
                    self._state_callbacks[key].remove(cb_tupl)
                    removed_count += 1
                if callback_id and callback_id == cb_id_:
                    self._state_callbacks[key].remove(cb_tupl)
                    removed_count += 1
                if len(self._state_callbacks[key]) == 0:
                    self._state_callbacks.pop(key)
        return removed_count

    def removeStateChangedCallbackByState(self, oldstate=None, newstate=None):
        """
        Removes all callback methods with given condition.

        This is useful if you use ``lambda x: do_somethig()`` to catch
        the moment when state gets changed.
        """
        for key in list(self._state_callbacks.keys()):
            if key == (
                oldstate,
                newstate,
            ):
                self._state_callbacks.pop(key)
                break

    def executeStateChangedCallbacks(self, oldstate, newstate, event_string, *args, **kwargs):
        """
        Compare conditions and execute state changed callback methods.
        """
        for key, cb_list in list(self._state_callbacks.items()) + list(self._callbacks_before_die.items()):
            old, new = key
            catched = False
            if old is None and new is None:
                catched = True
            elif old is None and new == newstate and newstate != oldstate:
                catched = True
            elif new is None and old == oldstate and newstate != oldstate:
                catched = True
            elif old == oldstate and new == newstate:
                catched = True
            if catched:
                for cb_tupl in cb_list:
                    _, cb = cb_tupl
                    cb(oldstate, newstate, event_string, *args, **kwargs)
        self._callbacks_before_die.clear()

    def publishEvents(self, on_off, publish_event_state_not_changed=None, publish_fast=None):
        """
        This can be used to enable "publishing" of all updates of the state machine to external "listeners".
        """
        self.publish_events = bool(on_off)
        if publish_event_state_not_changed is not None:
            self.publish_event_state_not_changed = publish_event_state_not_changed
        if publish_fast is not None:
            self.publish_fast = publish_fast

    def pushEvent(self, oldstate, newstate, event_string, publisher=None):
        """
        Can be used to
        """
        state_snapshot = dict(
            index=self.index,
            id=self.id,
            name=self.__class__.__name__,
            newstate=newstate,
            oldstate=oldstate,
            event=event_string,
        )
        if publisher is not None:
            if self.publish_fast:
                publisher(state_snapshot)
            else:
                reactor.callLater(0, publisher, state_snapshot)  # @UndefinedVariable
            return
        from bitdust.main import events
        events.send('state-changed', data=state_snapshot, fast=self.publish_fast)
